// DesignerViewStateViewport.cpp : Implementation of CDesignerViewStateViewport

#include "stdafx.h"
#include "DesignerViewStateViewport.h"


// CDesignerViewStateViewport

STDMETHODIMP CDesignerViewStateViewport::CLSIDGet(CLSID* a_pCLSID)
{
	try
	{
		*a_pCLSID = GetObjectCLSID();
		return S_OK;
	}
	catch (...)
	{
		return E_POINTER;
	}
}

STDMETHODIMP CDesignerViewStateViewport::ToText(BSTR* a_pbstrText)
{
	try
	{
		*a_pbstrText = NULL;
		OLECHAR sz[256];
		swprintf(sz, L"%f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f",
			m_tView._11, m_tView._12, m_tView._13, m_tView._14,
			m_tView._21, m_tView._22, m_tView._23, m_tView._24,
			m_tView._31, m_tView._32, m_tView._33, m_tView._34,
			m_tView._41, m_tView._42, m_tView._43, m_tView._44);
		*a_pbstrText = SysAllocString(sz);
		return S_OK;
	}
	catch (...)
	{
		return E_POINTER;
	}
}

STDMETHODIMP CDesignerViewStateViewport::FromText(BSTR a_bstrText)
{
	try
	{
		swscanf(a_bstrText, L"%f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f",
			&m_tView._11, &m_tView._12, &m_tView._13, &m_tView._14,
			&m_tView._21, &m_tView._22, &m_tView._23, &m_tView._24,
			&m_tView._31, &m_tView._32, &m_tView._33, &m_tView._34,
			&m_tView._41, &m_tView._42, &m_tView._43, &m_tView._44);
		return S_OK;
	}
	catch (...)
	{
		return E_UNEXPECTED;
	}
}

STDMETHODIMP CDesignerViewStateViewport::Init(ULONG a_nSizeX, ULONG a_nSizeY, TMatrix4x4f const* a_pProjection, TMatrix4x4f const* a_pView, ULONG a_nPointLights, TPointLight const* a_aPointLights, ULONG a_nDirectionalLights, TDirectionalLight const* a_aDirectionalLights)
{
	try
	{
		// TODO: lock
		m_nSizeX = a_nSizeX;
		m_nSizeY = a_nSizeY;

		m_tProj = a_pProjection != NULL ? *a_pProjection : TMATRIX4X4F_IDENTITY;
		m_tView = a_pView != NULL ? *a_pView : TMATRIX4X4F_IDENTITY;

		m_nPointLights = 0;
		delete[] m_aPointLights;
		m_aPointLights = NULL;
		m_nDirectionalLights = 0;
		delete[] m_aDirectionalLights;
		m_aDirectionalLights = NULL;

		if (a_nPointLights > 0)
		{
			m_aPointLights = new TPointLight[a_nPointLights];
			memcpy(m_aPointLights, a_aPointLights, sizeof(a_aPointLights[0])*a_nPointLights);
			m_nPointLights = a_nPointLights;
		}
		if (a_nDirectionalLights > 0)
		{
			m_aDirectionalLights = new TDirectionalLight[a_nDirectionalLights];
			memcpy(m_aDirectionalLights, a_aDirectionalLights, sizeof(a_aDirectionalLights[0])*a_nDirectionalLights);
			m_nDirectionalLights = a_nDirectionalLights;
		}

		return S_OK;
	}
	catch (...)
	{
		return E_UNEXPECTED;
	}
}

STDMETHODIMP CDesignerViewStateViewport::CameraGet(TMatrix4x4f* a_pProjection, TMatrix4x4f* a_pView)
{
	try
	{
		// TODO: lock
		if (a_pProjection)
		{
			*a_pProjection = m_tProj;
		}
		if (a_pView)
		{
			*a_pView = m_tView;
		}

		return S_OK;
	}
	catch (...)
	{
		return E_POINTER;
	}
}

STDMETHODIMP CDesignerViewStateViewport::LightsGet(IEnumStructs** a_ppPointLights, IEnumStructs** a_ppDirectionalLights)
{
	// TODO: lock
	HRESULT hRes = S_OK;

	try
	{
		*a_ppPointLights = NULL;
		CComPtr<IEnumStructsInit> pPLs;
		RWCoCreateInstance(pPLs, __uuidof(EnumStructs));
		pPLs->InsertMultiple(m_nPointLights, sizeof(TPointLight), reinterpret_cast<BYTE const*>(m_aPointLights));
		*a_ppPointLights = pPLs.Detach();
	}
	catch (...)
	{
		hRes = E_UNEXPECTED;
	}

	try
	{
		*a_ppDirectionalLights = NULL;
		CComPtr<IEnumStructsInit> pDLs;
		RWCoCreateInstance(pDLs, __uuidof(EnumStructs));
		pDLs->InsertMultiple(m_nDirectionalLights, sizeof(TDirectionalLight), reinterpret_cast<BYTE const*>(m_aDirectionalLights));
		*a_ppDirectionalLights = pDLs.Detach();
	}
	catch (...)
	{
		hRes = E_UNEXPECTED;
	}

	return hRes;
}

STDMETHODIMP CDesignerViewStateViewport::WindowGet(ULONG* a_pnSizeX, ULONG* a_pnSizeY)
{
	try
	{
		// TODO: lock
		*a_pnSizeX = m_nSizeX;
		*a_pnSizeY = m_nSizeY;

		return S_OK;
	}
	catch (...)
	{
		return E_POINTER;
	}
}

