// CustomNewFileTemplates.cpp : Implementation of CCustomNewFileTemplates

#include "stdafx.h"
#include "CustomNewFileTemplates.h"
#include <SharedStringTable.h>
#include <MultiLanguageString.h>
#include <shlobj.h>
#include <ApplicationFileName.h>


CCustomNewFileTemplates::CCustomNewFileTemplates() : m_bInitialized(false)
{
}

CCustomNewFileTemplates::~CCustomNewFileTemplates()
{
}

void CCustomNewFileTemplates::InternInit()
{
	if (!m_bInitialized)
	{
		CComPtr<ILocalizedString> pDesc;
		pDesc.Attach(_SharedStringTable.GetString(IDS_CUSTTEMP_FILEDESC));

		CComPtr<IApplicationInfo> pAI;
		RWCoCreateInstance(pAI, __uuidof(ApplicationInfo));
		CComBSTR bstrDir;
		if (pAI == NULL || FAILED(pAI->AppDataFolder(&bstrDir)) || bstrDir.Length() == 0)
			throw E_FAIL;
		bstrDir += L"\\Templates\\*.*";
		COLE2CT strDir(bstrDir.m_str);

		TCHAR szFile[MAX_PATH];
		_tcscpy(szFile, strDir);
		int iLen = _tcslen(szFile)-3; // subtract "*.*"
		WIN32_FIND_DATA cFindData;
		HANDLE hDir = FindFirstFile(strDir, &cFindData);
		if (hDir != INVALID_HANDLE_VALUE)
		{
			USES_CONVERSION;

			do
			{
				if (_tcscmp(cFindData.cFileName, _T(".")) != 0 &&
					_tcscmp(cFindData.cFileName, _T("..")) != 0)
				{
					SItem sItem;
					_tcscpy(szFile + iLen, cFindData.cFileName);
					sItem.bstrPath = szFile;
					CComPtr<ILocalizedStringInit> pInit;
					RWCoCreateInstance(pInit, __uuidof(LocalizedString));
					pInit->Insert(MAKELCID(MAKELANGID(LANG_NEUTRAL, SUBLANG_NEUTRAL), SORT_DEFAULT), CComBSTR(cFindData.cFileName));
					sItem.pName.Attach(pInit.Detach());
					sItem.pDesc = pDesc;
					m_cItems.push_back(sItem);
				}
			} while(FindNextFile(hDir, &cFindData));
			FindClose(hDir);
		}
		m_bInitialized = true;
	}
}

HRESULT CCustomNewFileTemplates::TemplatesSize(ULONG* a_pnCount)
{
	try
	{
		ObjectLock cLock(this);
		InternInit();

		*a_pnCount = m_cItems.size();//+1;
		return S_OK;
	}
	catch (...)
	{
		return a_pnCount == NULL ? E_POINTER : E_UNEXPECTED;
	}
}

HRESULT CCustomNewFileTemplates::TemplatesName(ULONG a_nIndex, ILocalizedString** a_ppName)
{
	try
	{
		*a_ppName = NULL;
		ObjectLock cLock(this);
		InternInit();

		if (a_nIndex == m_cItems.size())
		{
			*a_ppName = new CMultiLanguageString(L"[0409]Add custom template[0405]Pidat vlastn ablonu");
			return S_OK;
		}

		if (a_nIndex >= m_cItems.size())
			return E_RW_INDEXOUTOFRANGE;

		SItem& sItem = m_cItems[a_nIndex];
		(*a_ppName = sItem.pName)->AddRef();

		return S_OK;
	}
	catch (...)
	{
		return a_ppName == NULL ? E_POINTER : E_UNEXPECTED;
	}
}

HRESULT CCustomNewFileTemplates::TemplatesHelpText(ULONG a_nIndex, ILocalizedString** a_ppHelpText)
{
	try
	{
		*a_ppHelpText = NULL;
		ObjectLock cLock(this);
		InternInit();

		if (a_nIndex == m_cItems.size())
		{
			*a_ppHelpText = new CMultiLanguageString(L"[0409]Make a template from a file. Copy of the file is opened when the template is used.[0405]Udlat ablonu ze souboru. Kopie souboru bude otevena pi pouit ablony.");
			return S_OK;
		}

		if (a_nIndex >= m_cItems.size())
			return E_RW_INDEXOUTOFRANGE;

		SItem& sItem = m_cItems[a_nIndex];
		(*a_ppHelpText = sItem.pDesc)->AddRef();

		return S_OK;
	}
	catch (...)
	{
		return a_ppHelpText == NULL ? E_POINTER : E_UNEXPECTED;
	}
}

HRESULT CCustomNewFileTemplates::TemplatesIcon(ULONG a_nIndex, ULONG a_nSize, HICON* a_phIcon)
{
	try
	{
		*a_phIcon = NULL;
		ObjectLock cLock(this);
		InternInit();

		if (a_nIndex == m_cItems.size())
		{
			*a_phIcon = (HICON)LoadImage(_pModule->get_m_hInst(), MAKEINTRESOURCE(IDI_CUSTOMTEMPLATES), IMAGE_ICON, a_nSize, a_nSize, LR_DEFAULTCOLOR);
			return S_OK;
		}

		if (a_nIndex >= m_cItems.size())
			return E_RW_INDEXOUTOFRANGE;

		SItem& sItem = m_cItems[a_nIndex];
		SHFILEINFO tInfo;
		ZeroMemory(&tInfo, sizeof tInfo);
		USES_CONVERSION;
		SHGetFileInfo(OLE2CT(sItem.bstrPath), 0, &tInfo, sizeof tInfo, SHGFI_ICON|SHGFI_LARGEICON/*|SHGFI_SMALLICON*/);
		*a_phIcon = tInfo.hIcon;

		return S_OK;
	}
	catch (...)
	{
		return a_phIcon == NULL ? E_POINTER : E_UNEXPECTED;
	}
}

STDMETHODIMP CCustomNewFileTemplates::Category(ULONG a_nIndex, ILocalizedString** a_ppCategory)
{
	try
	{
		*a_ppCategory = 0;
		if (a_nIndex > m_cItems.size())
			return E_RW_INDEXOUTOFRANGE;

		*a_ppCategory = _SharedStringTable.GetString(IDS_CATEGORY_CUSTOM);
		return S_OK;
	}
	catch (...)
	{
		return a_ppCategory == NULL ? E_POINTER : E_UNEXPECTED;
	}
}

STDMETHODIMP CCustomNewFileTemplates::CheckFeatures(ULONG a_nIndex, ULONG a_nCount, const IID* a_aiidRequired)
{
	try
	{
		if (a_nIndex > m_cItems.size())
			return E_RW_INDEXOUTOFRANGE;

		return S_OK; //TODO: implement correctly
	}
	catch (...)
	{
		return E_UNEXPECTED;
	}
}

STDMETHODIMP CCustomNewFileTemplates::DocumentCreate(ULONG a_nIndex, RWHWND UNREF(a_hParentWnd), LCID UNREF(a_tLocaleID), IConfig* UNREF(a_pConfig), BSTR a_bstrPrefix, IDocumentBase* a_pBase)
{
	try
	{
		ObjectLock cLock(this);
		InternInit();

		CComPtr<IInputManager> pManager;
		RWCoCreateInstance(pManager, __uuidof(InputManager));

		if (a_nIndex >= m_cItems.size())
			return E_RW_INDEXOUTOFRANGE;

		return pManager->DocumentCreateData(0, NULL, CStorageFilter(m_cItems[a_nIndex].bstrPath), a_bstrPrefix, a_pBase);
	}
	catch (...)
	{
		return E_UNEXPECTED;
	}
}

HRESULT CCustomNewFileTemplates::ToolsSize(ULONG* a_pnCount)
{
	try
	{
		*a_pnCount = 1;
		return S_OK;
	}
	catch (...)
	{
		return E_POINTER;
	}
}

HRESULT CCustomNewFileTemplates::ToolsName(ULONG a_nIndex, ILocalizedString** a_ppName)
{
	try
	{
		*a_ppName = NULL;
		if (a_nIndex != 0)
			return E_RW_INDEXOUTOFRANGE;
		*a_ppName = _SharedStringTable.GetString(IDS_CUSTTEMP_MANAGEUSER_NAME);
		return S_OK;
	}
	catch (...)
	{
		return a_ppName == NULL ? E_POINTER : E_UNEXPECTED;
	}
}

HRESULT CCustomNewFileTemplates::ToolsHelpText(ULONG a_nIndex, ILocalizedString** a_ppHelpText)
{
	try
	{
		*a_ppHelpText = NULL;
		if (a_nIndex != 0)
			return E_RW_INDEXOUTOFRANGE;
		*a_ppHelpText = _SharedStringTable.GetString(IDS_CUSTTEMP_MANAGEUSER_DESC);
		return S_OK;
	}
	catch (...)
	{
		return a_ppHelpText == NULL ? E_POINTER : E_UNEXPECTED;
	}
}

HRESULT CCustomNewFileTemplates::ToolsIcon(ULONG a_nIndex, ULONG a_nSize, HICON* a_phIcon)
{
	try
	{
		*a_phIcon = NULL;
		return E_NOTIMPL;
	}
	catch (...)
	{
		return a_phIcon == NULL ? E_POINTER : E_UNEXPECTED;
	}
}

STDMETHODIMP CCustomNewFileTemplates::Activate(ULONG a_nIndex, RWHWND a_hFrameWnd, LCID a_tLocaleID, IDocument* a_pDocument)
{
	try
	{
		if (a_nIndex != 0)
			return E_RW_INDEXOUTOFRANGE;
		MessageBox(reinterpret_cast<HWND>(a_hFrameWnd), _T("Manage custom teplates"), _T("Information"), MB_OK);
		return S_OK;
	}
	catch (...)
	{
		return E_UNEXPECTED;
	}
}

