
#pragma once


#include "EditTool.h"
#include "EditToolScanlineBuffer.h"
#include "EditToolWithBrush.h"
#include "EditToolRectangularShape.h"

#include <agg_scanline_p.h>
#include <agg_renderer_scanline.h>
#include <agg_pixfmt_rgba.h>
#include <agg_rasterizer_scanline_aa.h>
#include <agg_rounded_rect.h>
#include <agg_trans_affine.h>
#include <agg_rasterizer_compound_aa.h>
#include <agg_span_allocator.h>
#include <agg_scanline_u.h>
#include <agg_span_gradient.h>
#include <agg_conv_stroke.h>
#include <agg_conv_transform.h>


#include <agg_basics.h>
#include <agg_color_rgba.h>
#include <agg_pixfmt_rgba.h>
#include <agg_rendering_buffer.h>
#include <agg_scanline_p.h>
#include <agg_scanline_u.h>
#include <agg_trans_perspective.h>
#include <agg_trans_bilinear.h>
#include <agg_renderer_scanline.h>
#include <agg_rasterizer_scanline_aa.h>
#include <agg_span_subdiv_adaptor.h>
#include <agg_span_allocator.h>
//#include <agg_span_image_resample_rgba.h>
#include <agg_span_image_filter_rgba.h>
#include <agg_span_interpolator_trans.h>
#include <agg_image_accessors.h>

struct CEditToolDataTransformation
{
	MIDL_INTERFACE("DC497588-B34E-4C3C-928D-179D8AEAAB6E")
	ISharedStateToolData : public ISharedState
	{
	public:
		STDMETHOD_(CEditToolDataTransformation const*, InternalData)() = 0;
	};
	enum EMode
	{
		EMSimple = 0,
		EMPerspective,
		EMBezier,
	};

	CEditToolDataTransformation() : eMode(EMSimple), bAspectLock(false)
	{
	}

	HRESULT FromString(BSTR a_bstr)
	{
		if (a_bstr == NULL)
			return S_OK;
		if (wcsstr(a_bstr, L"SIMPLE"))
			eMode = EMSimple;
		else if (wcsstr(a_bstr, L"PERSPECTIVE"))
			eMode = EMPerspective;
		else if (wcsstr(a_bstr, L"BEZIER"))
			eMode = EMBezier;
		bAspectLock = wcsstr(a_bstr, L"LOCK");
		return S_OK;
	}
	HRESULT ToString(BSTR* a_pbstr)
	{
		*a_pbstr = SysAllocString(eMode == EMSimple ? (bAspectLock ? L"SIMPLE|LOCK" : L"SIMPLE") : (eMode == EMPerspective ? (bAspectLock ? L"PERSPECTIVE|LOCK" : L"PERSPECTIVE") : (bAspectLock ? L"BEZIER|LOCK" : L"BEZIER")));
		return S_OK;
	}

	EMode eMode;
	bool bAspectLock;
};

#include "EditToolTransformDlg.h"


class CEditToolTransformation :
	public CEditToolRectangularShape<CEditToolTransformation>, // creating and modifying the rectangle
	public CEditToolMouseInput<CEditToolTransformation, CEditToolRectangularShape<CEditToolTransformation> >, // no direct tablet support
	public CEditToolCustomOrMoveCursor<CEditToolTransformation, IDC_EDITTOOL_TRANSFORMATION>, // cursor handler
	public CRasterImageEditToolImpl< // IRasterImageEditTool implementation mapper
		CEditToolTransformation, // T - the top level class for cross casting
		CEditToolTransformation, // TResetHandler
		CEditToolTransformation, // TDirtyHandler
		CEditToolTransformation, // TImageTileHandler
		CEditToolTransformation, // TSelectionTileHandler
		CRasterImageEditToolBase, // TColorsHandler
		CRasterImageEditToolBase, // TBrushHandler
		CEditToolTransformation, // TGlobalsHandler
		CEditToolTransformation, // TAdjustCoordsHandler
		CEditToolCustomOrMoveCursor<CEditToolTransformation, IDC_EDITTOOL_TRANSFORMATION>, // TGetCursorHandler
		CEditToolMouseInput<CEditToolTransformation, CEditToolRectangularShape<CEditToolTransformation> >, // TProcessInputHandler
		CEditToolTransformation, // TPreTranslateMessageHandler
		CEditToolRectangularShape<CEditToolTransformation>, // TControlPointsHandler
		CEditToolRectangularShape<CEditToolTransformation> // TControlLinesHandler
	>,
	public IRasterImageEditToolFloatingSelection,
	public IRasterImageEditToolScripting,
	public IRasterImageEditToolContextMenu,
	public IDesignerViewStatusBar
{
public:
	CEditToolTransformation() :
		m_eBlendingMode(EBMDrawOver), m_eRasterizationMode(ERMSmooth), m_eCoordinatesMode(ECMFloatingPoint),
		CEditToolRectangularShape<CEditToolTransformation>(CEditToolRectangularShape::ERMRotatedRectangle, true, false),
		m_rcPrev(RECT_EMPTY), m_rcDelete(RECT_EMPTY), m_bUseDeleteMask(false)
	{
		m_szSelSize.cx = m_szSelSize.cy = 0;
	}

	BEGIN_COM_MAP(CEditToolTransformation)
		COM_INTERFACE_ENTRY(IRasterImageEditTool)
		COM_INTERFACE_ENTRY(IRasterImageEditToolScripting)
		COM_INTERFACE_ENTRY(IRasterImageEditToolFloatingSelection)
		COM_INTERFACE_ENTRY(IRasterImageEditToolContextMenu)
		COM_INTERFACE_ENTRY(IDesignerViewStatusBar)
	END_COM_MAP()

	// IRasterImageEditToolContextMenu methods
public:
	class ATL_NO_VTABLE CRestoreSelection :
		public CDocumentMenuCommandImpl<CRestoreSelection, IDS_MENU_RESTORESELECTION_NAME, IDS_MENU_RESTORESELECTION_DESC, NULL, 0>
	{
	public:
		CRestoreSelection() : m_pTool(NULL)
		{
		}
		~CRestoreSelection()
		{
			if (m_pTool) m_pTool->Release();
		}
		void Init(CEditToolTransformation* a_pTool)
		{
			(m_pTool = a_pTool)->AddRef();
		}

		// IDocumentMenuCommand
	public:
		STDMETHOD(Execute)(RWHWND a_hParent, LCID a_tLocaleID)
		{
			return m_pTool->RestoreSelection();
		}

	private:
		CEditToolTransformation* m_pTool;
	};
	class ATL_NO_VTABLE CStretchSelection :
		public CDocumentMenuCommandImpl<CStretchSelection, IDS_MENU_STRETCHSELECTION_NAME, IDS_MENU_STRETCHSELECTION_DESC, NULL, 0>
	{
	public:
		CStretchSelection() : m_pTool(NULL)
		{
		}
		~CStretchSelection()
		{
			if (m_pTool) m_pTool->Release();
		}
		void Init(CEditToolTransformation* a_pTool)
		{
			(m_pTool = a_pTool)->AddRef();
		}

		// IDocumentMenuCommand
	public:
		STDMETHOD(Execute)(RWHWND a_hParent, LCID a_tLocaleID)
		{
			return m_pTool->StretchSelection();
		}

	private:
		CEditToolTransformation* m_pTool;
	};
	class ATL_NO_VTABLE CCropToSelection :
		public CDocumentMenuCommandImpl<CCropToSelection, IDS_MENU_CROP2SELECTION_NAME, IDS_MENU_CROP2SELECTION_DESC, NULL, 0>
	{
	public:
		CCropToSelection() : m_pTool(NULL)
		{
		}
		~CCropToSelection()
		{
			if (m_pTool) m_pTool->Release();
		}
		void Init(CEditToolTransformation* a_pTool)
		{
			(m_pTool = a_pTool)->AddRef();
		}

		// IDocumentMenuCommand
	public:
		STDMETHOD(Execute)(RWHWND a_hParent, LCID a_tLocaleID)
		{
			return m_pTool->CropToSelection();
		}

	private:
		CEditToolTransformation* m_pTool;
	};

	HRESULT RestoreSelection()
	{
		if (M_Window() == NULL || M_State() == ESClean)
			return S_FALSE;
		bool bOrigPos = m_rcDelete.left < m_rcDelete.right && m_rcDelete.top < m_rcDelete.bottom;
		if (m_cData.eMode == CEditToolDataTransformation::EMPerspective)
			if (bOrigPos)
				SetRectanglePoints(m_rcDelete.left, m_rcDelete.top, m_rcDelete.right, m_rcDelete.top, m_rcDelete.right, m_rcDelete.bottom, m_rcDelete.left, m_rcDelete.bottom);
			else
				SetRectanglePoints(0.0f, 0.0f, m_szSelSize.cx, 0.0f, m_szSelSize.cx, m_szSelSize.cy, 0.0f, m_szSelSize.cy);
		else
			SetRectangle(bOrigPos ? (m_rcDelete.left+m_rcDelete.right)*0.5f : m_szSelSize.cx*0.5f, bOrigPos ? (m_rcDelete.top+m_rcDelete.bottom)*0.5f : m_szSelSize.cy*0.5f, m_szSelSize.cx*0.5f, m_szSelSize.cy*0.5f, 0.0f);
		return S_OK;
	}
	HRESULT StretchSelection()
	{
		if (M_Window() == NULL || M_State() == ESClean)
			return S_FALSE;
		if (m_cData.eMode == CEditToolDataTransformation::EMPerspective)
			SetRectanglePoints(0.0f, 0.0f, m_nSizeX, 0.0f, m_nSizeX, m_nSizeY, 0.0f, m_nSizeY);
		else
			SetRectangle(m_nSizeX*0.5f, m_nSizeY*0.5f, m_nSizeX*0.5f, m_nSizeY*0.5f, 0.0f);
		return S_OK;
	}
	HRESULT CropToSelection()
	{
		if (M_Window() == NULL || M_State() == ESClean)
			return S_FALSE;
		CComPtr<IDocument> pDoc;
		M_Window()->Document(&pDoc);
		CComPtr<IDocumentRasterImage> pRI;
		pDoc->QueryFeatureInterface(__uuidof(IDocumentRasterImage), reinterpret_cast<void**>(&pRI));
		if (pRI)
		{
			// simple rectangular crop
			TImageSize tSize = {m_szSelSize.cx, m_szSelSize.cy};
			CWriteLock<IDocument> cLock(pDoc);
			RECT const rc = GetBoundingBox();
			TMatrix3x3f const tMtx = {1.0f, 0.0f, 0.0f,  0.0f, 1.0f, 0.0f,  -rc.left, -rc.top, 1.0f};
			pRI->CanvasSet(&tSize, NULL, &tMtx, NULL);
			pRI->TileSet(EICIRGBA, NULL, NULL, NULL, tSize.nX*tSize.nY, reinterpret_cast<TPixelChannel*>(m_cSelData.m_p), TRUE);
		}
		return S_OK;
	}

	STDMETHOD(CommandsEnum)(TPixelCoords const* a_pPixel, ULONG a_nControlPointIndex, IEnumUnknowns** a_ppSubCommands)
	{
		try
		{
			*a_ppSubCommands = NULL;

			if (_IsDirty(NULL, NULL, NULL) != S_OK)
				return S_FALSE;

			CComPtr<IEnumUnknownsInit> pItems;
			RWCoCreateInstance(pItems, __uuidof(EnumUnknowns));

			{
				CComObject<CRestoreSelection>* p = NULL;
				CComObject<CRestoreSelection>::CreateInstance(&p);
				CComPtr<IDocumentMenuCommand> pTmp = p;
				p->Init(this);
				pItems->Insert(pTmp);
			}
			{
				CComObject<CStretchSelection>* p = NULL;
				CComObject<CStretchSelection>::CreateInstance(&p);
				CComPtr<IDocumentMenuCommand> pTmp = p;
				p->Init(this);
				pItems->Insert(pTmp);
			}
			{
				CComObject<CCropToSelection>* p = NULL;
				CComObject<CCropToSelection>::CreateInstance(&p);
				CComPtr<IDocumentMenuCommand> pTmp = p;
				p->Init(this);
				pItems->Insert(pTmp);
			}

			*a_ppSubCommands = pItems.Detach();
			return S_OK;
		}
		catch (...)
		{
			return a_ppSubCommands ? E_UNEXPECTED : E_POINTER;
		}
	}

	// IDesignerViewStatusBar
public:
	STDMETHOD(Update)(IDesignerStatusBar* a_pStatusBar)
	{
		OLECHAR szTmp[64];
		if (m_cSelData.m_p)
		{
			swprintf(szTmp, L"%ix%i", m_szSelSize.cx, m_szSelSize.cy);
		}
		else if (M_State() == ESDragging)
		{
			RECT rcBounds = GetBoundingBox();
			if (rcBounds.right-rcBounds.left || rcBounds.bottom-rcBounds.top)
				swprintf(szTmp, L"%ix%i", rcBounds.right-rcBounds.left, rcBounds.bottom-rcBounds.top);
			else
				return S_FALSE;
		}
		else
			return S_FALSE;
		return a_pStatusBar->PaneSet(CComBSTR("TRANSFORMSIZE"), NULL, CComBSTR(szTmp), 60, -200);
	}



	ECoordinatesMode M_CoordinatesMode() const
	{
		return m_eCoordinatesMode;
	}

	template<class TPixelMixer, class TTransform>
	void DrawSimple(RECT const& a_rc, RECT const& a_rcDirty, TRasterImagePixel* a_pBuffer, ULONG a_nStride, TTransform a_tr)
	{
		for (LONG y = a_rc.top; y < a_rc.bottom; ++y)
		{
			TRasterImagePixel* pBuf = a_pBuffer+(y-a_rcDirty.top)*a_nStride+a_rc.left-a_rcDirty.left;
			for (LONG x = a_rc.left; x < a_rc.right; ++x, ++pBuf)
			{
				double xx = x+0.5;
				double yy = y+0.5;
				a_tr.transform(&xx, &yy);
				LONG xxx = LONG(xx+1.0)-1;
				LONG yyy = LONG(yy+1.0)-1;
				if (xxx >= 0 && xxx < m_szSelSize.cx && yyy >= 0 && yyy < m_szSelSize.cy)
				{
					if (m_bUseDeleteMask)
						TPixelMixer::Mix(*pBuf, m_cSelData[yyy*m_szSelSize.cx+xxx], m_cDeleteMask[yyy*m_szSelSize.cx+xxx]);
					else
						TPixelMixer::Mix(*pBuf, m_cSelData[yyy*m_szSelSize.cx+xxx]);
				}
			}
		}
	}

	template<class TPixelMixer, class TTransform>
	void DrawSmooth(RECT const& a_rc, RECT const& a_rcDirty, TRasterImagePixel* a_pBuffer, ULONG a_nStride, TTransform a_tr)
	{
		for (LONG y = a_rc.top; y < a_rc.bottom; ++y)
		{
			TRasterImagePixel* pBuf = a_pBuffer+(y-a_rcDirty.top)*a_nStride+a_rc.left-a_rcDirty.left;
			for (LONG x = a_rc.left; x < a_rc.right; ++x, ++pBuf)
			{
				double xx = x+0.5;
				double yy = y+0.5;
				a_tr.transform(&xx, &yy);
				LONG xxx = LONG(xx+1.5)-2;
				LONG yyy = LONG(yy+1.5)-2;
				if (xxx >= -1 && xxx < m_szSelSize.cx && yyy >= -1 && yyy < m_szSelSize.cy)
				{
					ULONG nWX = 128*(xx-xxx-0.5);
					ULONG nWY = 128*(yy-yyy-0.5);
					ULONG nR = 0;
					ULONG nG = 0;
					ULONG nB = 0;
					ULONG nA = 0;
					ULONG nC = 0;
					TRasterImagePixel* p;
					if (xxx >= 0)
					{
						if (xxx != m_szSelSize.cx-1)
						{
							if (yyy >= 0)
							{
								if (yyy != m_szSelSize.cy-1)
								{
									{
										p = m_cSelData.m_p + yyy*m_szSelSize.cx+xxx;
										ULONG const nW = (128-nWY)*(128-nWX);
										ULONG const n = nW*p->bA;
										nA += n;
										nR += n*p->bR;
										nG += n*p->bG;
										nB += n*p->bB;
										nC += m_bUseDeleteMask ? m_cDeleteMask[yyy*m_szSelSize.cx+xxx]*nW : 255*nW;
									}
									{
										p = m_cSelData.m_p + (yyy+1)*m_szSelSize.cx+xxx;
										ULONG const nW = nWY*(128-nWX);
										ULONG const n = nW*p->bA;
										nA += n;
										nR += n*p->bR;
										nG += n*p->bG;
										nB += n*p->bB;
										nC += m_bUseDeleteMask ? m_cDeleteMask[(yyy+1)*m_szSelSize.cx+xxx]*nW : 255*nW;
									}
									{
										p = m_cSelData.m_p + yyy*m_szSelSize.cx+xxx+1;
										ULONG const nW = (128-nWY)*nWX;
										ULONG const n = nW*p->bA;
										nA += n;
										nR += n*p->bR;
										nG += n*p->bG;
										nB += n*p->bB;
										nC += m_bUseDeleteMask ? m_cDeleteMask[yyy*m_szSelSize.cx+xxx+1]*nW : 255*nW;
									}
									{
										p = m_cSelData.m_p + (yyy+1)*m_szSelSize.cx+xxx+1;
										ULONG const nW = nWY*nWX;
										ULONG const n = nW*p->bA;
										nA += n;
										nR += n*p->bR;
										nG += n*p->bG;
										nB += n*p->bB;
										nC += m_bUseDeleteMask ? m_cDeleteMask[(yyy+1)*m_szSelSize.cx+xxx+1]*nW : 255*nW;
									}
								}
								else
								{
									{
										p = m_cSelData.m_p + yyy*m_szSelSize.cx+xxx;
										ULONG const n = ((128-nWX)<<7)*p->bA;
										nA += n;
										nR += n*p->bR;
										nG += n*p->bG;
										nB += n*p->bB;
										nC += m_bUseDeleteMask ? m_cDeleteMask[(yyy+1)*m_szSelSize.cx+xxx]*(128-nWX)*(128-nWY) : 255*(128-nWX)*(128-nWY);
									}
									{
										p = m_cSelData.m_p + yyy*m_szSelSize.cx+xxx+1;
										ULONG const n = (nWX<<7)*p->bA;
										nA += n;
										nR += n*p->bR;
										nG += n*p->bG;
										nB += n*p->bB;
										nC += m_bUseDeleteMask ? m_cDeleteMask[yyy*m_szSelSize.cx+xxx+1]*nWX*(128-nWY) : 255*nWX*(128-nWY);
									}
								}
							}
							else
							{
								{
									p = m_cSelData.m_p + (yyy+1)*m_szSelSize.cx+xxx;
									ULONG const n = ((128-nWX)<<7)*p->bA;
									nA += n;
									nR += n*p->bR;
									nG += n*p->bG;
									nB += n*p->bB;
									nC += m_bUseDeleteMask ? m_cDeleteMask[(yyy+1)*m_szSelSize.cx+xxx]*(128-nWX)*nWY : 255*(128-nWX)*nWY;
								}
								{
									p = m_cSelData.m_p + (yyy+1)*m_szSelSize.cx+xxx+1;
									ULONG const n = (nWX<<7)*p->bA;
									nA += n;
									nR += n*p->bR;
									nG += n*p->bG;
									nB += n*p->bB;
									nC += m_bUseDeleteMask ? m_cDeleteMask[(yyy+1)*m_szSelSize.cx+xxx+1]*nWX*nWY : 255*nWX*nWY;
								}
							}
						}
						else
						{
							if (yyy >= 0)
							{
								if (yyy != m_szSelSize.cy-1)
								{
									{
										p = m_cSelData.m_p + yyy*m_szSelSize.cx+xxx;
										ULONG const n = ((128-nWY)<<7)*p->bA;
										nA += n;
										nR += n*p->bR;
										nG += n*p->bG;
										nB += n*p->bB;
										nC += m_bUseDeleteMask ? m_cDeleteMask[yyy*m_szSelSize.cx+xxx]*(128-nWY)*(128-nWX) : 255*(128-nWY)*(128-nWX);
									}
									{
										p = m_cSelData.m_p + (yyy+1)*m_szSelSize.cx+xxx;
										ULONG const n = (nWY<<7)*p->bA;
										nA += n;
										nR += n*p->bR;
										nG += n*p->bG;
										nB += n*p->bB;
										nC += m_bUseDeleteMask ? m_cDeleteMask[(yyy+1)*m_szSelSize.cx+xxx]*nWY*(128-nWX) : 255*nWY*(128-nWX);
									}
								}
								else
								{
									p = m_cSelData.m_p + yyy*m_szSelSize.cx+xxx;
									ULONG const n = ULONG(p->bA)<<14;
									nA += n;
									nR += n*p->bR;
									nG += n*p->bG;
									nB += n*p->bB;
									nC += m_bUseDeleteMask ? m_cDeleteMask[yyy*m_szSelSize.cx+xxx]*(128-nWX)*(128-nWY) : 255*(128-nWX)*(128-nWY);
								}
							}
							else
							{
								p = m_cSelData.m_p + (yyy+1)*m_szSelSize.cx+xxx;
								ULONG const n = ULONG(p->bA)<<14;
								nA += n;
								nR += n*p->bR;
								nG += n*p->bG;
								nB += n*p->bB;
								nC += m_bUseDeleteMask ? m_cDeleteMask[(yyy+1)*m_szSelSize.cx+xxx]*(128-nWX)*nWY : 255*(128-nWX)*nWY;
							}
						}
					}
					else
					{
						if (yyy >= 0)
						{
							if (yyy != m_szSelSize.cy-1)
							{
								{
									p = m_cSelData.m_p + yyy*m_szSelSize.cx+xxx+1;
									ULONG const n = ((128-nWY)<<7)*p->bA;
									nA += n;
									nR += n*p->bR;
									nG += n*p->bG;
									nB += n*p->bB;
									nC += m_bUseDeleteMask ? m_cDeleteMask[yyy*m_szSelSize.cx+xxx]*(128-nWY)*nWX : 255*(128-nWY)*nWX;
								}
								{
									p = m_cSelData.m_p + (yyy+1)*m_szSelSize.cx+xxx+1;
									ULONG const n = (nWY<<7)*p->bA;
									nA += n;
									nR += n*p->bR;
									nG += n*p->bG;
									nB += n*p->bB;
									nC += m_bUseDeleteMask ? m_cDeleteMask[(yyy+1)*m_szSelSize.cx+xxx]*nWY*nWX : 255*nWY*nWX;
								}
							}
							else
							{
								p = m_cSelData.m_p + yyy*m_szSelSize.cx+xxx+1;
								ULONG const n = ULONG(p->bA)<<14;
								nA += n;
								nR += n*p->bR;
								nG += n*p->bG;
								nB += n*p->bB;
								nC += m_bUseDeleteMask ? m_cDeleteMask[yyy*m_szSelSize.cx+xxx+1]*nWX*(128-nWY) : 255*nWX*(128-nWY);
							}
						}
						else
						{
							p = m_cSelData.m_p + (yyy+1)*m_szSelSize.cx+xxx+1;
							ULONG const n = ULONG(p->bA)<<14;
							nA += n;
							nR += n*p->bR;
							nG += n*p->bG;
							nB += n*p->bB;
							nC += m_bUseDeleteMask ? m_cDeleteMask[(yyy+1)*m_szSelSize.cx+xxx+1]*nWX*nWY : 255*nWX*nWY;
						}
					}
					TRasterImagePixel t = {0, 0, 0, nA>>14};
					if (t.bA)
					{
						t.bR = nR>>22;
						t.bG = nG>>22;
						t.bB = nB>>22;
					}
					TPixelMixer::MixPM(*pBuf, t, nC>>14);
				}
			}
		}
	}

	// IRasterImageEditTool methods
public:
	HRESULT _Reset()
	{
		if (m_rcDelete.left < m_rcDelete.right && m_rcDelete.top < m_rcDelete.bottom)
			M_Window()->RectangleChanged(&m_rcDelete);
		if (m_rcPrev.left < m_rcPrev.right && m_rcPrev.top < m_rcPrev.bottom)
			M_Window()->RectangleChanged(&m_rcPrev);
		ResetRectangle();
		m_nSizeX = 0;
		m_nSizeY = 0;
		M_Window()->Size(&m_nSizeX, &m_nSizeY);
		m_cSelData.Free();
		m_cDeleteMask.Free();
		m_rcDelete = RECT_EMPTY;
		m_bUseDeleteMask = false;
		if (m_szSelSize.cx == 0 && m_szSelSize.cy == 0)
		{
			RECT rcBounds = {0, 0, m_nSizeX, m_nSizeY};
			BOOL bEntire = TRUE;
			M_Window()->GetSelectionInfo(&rcBounds, &bEntire);
			if ((rcBounds.left != 0 || rcBounds.top != 0 ||
				rcBounds.right != m_nSizeX || rcBounds.bottom != m_nSizeY ||
				bEntire != TRUE) && rcBounds.left < rcBounds.right && rcBounds.top < rcBounds.bottom)
			{
				SetRectangle((rcBounds.left+rcBounds.right)*0.5f, (rcBounds.top+rcBounds.bottom)*0.5f, (rcBounds.right-rcBounds.left)*0.5f, (rcBounds.bottom-rcBounds.top)*0.5f, 0.0f);
				m_rcPrev = GetBoundingBox();
				m_szSelSize.cx = rcBounds.right-rcBounds.left;
				m_szSelSize.cy = rcBounds.bottom-rcBounds.top;
				m_cSelData.Allocate(m_szSelSize.cx*m_szSelSize.cy);
				if (!bEntire)
				{
					m_cDeleteMask.Allocate(m_szSelSize.cx*m_szSelSize.cy);
					M_Window()->GetSelectionTile(rcBounds.left, rcBounds.top, rcBounds.right-rcBounds.left, rcBounds.bottom-rcBounds.top, rcBounds.right-rcBounds.left, m_cDeleteMask);
					m_bUseDeleteMask = true;
				}
				M_Window()->GetImageTile(rcBounds.left, rcBounds.top, rcBounds.right-rcBounds.left, rcBounds.bottom-rcBounds.top, 0.0f, rcBounds.right-rcBounds.left, EITIContent, m_cSelData);
				m_rcDelete = rcBounds;
				RECT rcAll = {0, 0, m_nSizeX, m_nSizeY};
				M_Window()->RectangleChanged(&rcAll);
			}
		}
		else
		{
			m_szSelSize.cx = m_szSelSize.cy = 0;
		}
		M_Window()->ControlPointsChanged();
		M_Window()->ControlLinesChanged();
		return S_OK;
	}

	HRESULT _GetImageTile(LONG a_nX, LONG a_nY, ULONG a_nSizeX, ULONG a_nSizeY, float a_fGamma, ULONG a_nStride, TRasterImagePixel* a_pBuffer)
	{
		HRESULT hRes = M_Window()->GetImageTile(a_nX, a_nY, a_nSizeX, a_nSizeY, a_fGamma, a_nStride, EITIBackground, a_pBuffer);
		if (FAILED(hRes))
			return hRes;
		RECT rcDirty = {a_nX, a_nY, a_nX+a_nSizeX, a_nY+a_nSizeY};
		RECT rc =
		{
			max(rcDirty.left, m_rcDelete.left),
			max(rcDirty.top, m_rcDelete.top),
			min(rcDirty.right, m_rcDelete.right),
			min(rcDirty.bottom, m_rcDelete.bottom),
		};
		if (rc.left < rc.right && rc.top < rc.bottom)
		{
			for (LONG y = rc.top; y < rc.bottom; ++y)
			{
				if (m_cDeleteMask.m_p)
				{
					TRasterImagePixel* pBuf = a_pBuffer+(y-rcDirty.top)*a_nStride+rc.left-rcDirty.left;
					BYTE* pDel = m_cDeleteMask.m_p+(y-m_rcDelete.top)*(m_rcDelete.right-m_rcDelete.left)+rc.left-m_rcDelete.left;
					for (LONG x = rc.left; x < rc.right; ++x)
					{
						pBuf->bA = pBuf->bA*(255-*pDel)/255;
						if (pBuf->bA == 0)
							pBuf->bR = pBuf->bG = pBuf->bB = 0;
						++pBuf;
						++pDel;
					}
				}
				else
				{
					ZeroMemory(a_pBuffer+(y-rcDirty.top)*a_nStride+rc.left-rcDirty.left, (rc.right-rc.left)*sizeof*a_pBuffer);
				}
			}
		}
		if (m_cSelData.m_p)
		{
			TPixelCoords const* pPts = GetRectanglePoints();

			if (pPts[0].fX == pPts[3].fX && pPts[1].fX == pPts[2].fX &&
				pPts[0].fY == pPts[1].fY && pPts[2].fY == pPts[3].fY &&
				pPts[0].fX == int(pPts[0].fX) && pPts[0].fY == int(pPts[0].fY) &&
				pPts[1].fX-pPts[0].fX == m_szSelSize.cx && pPts[3].fY-pPts[0].fY == m_szSelSize.cy)
			{
				// rectangle was only moved (no resizing, no rotation) - use optimized code
				rc.left = max(rcDirty.left, LONG(pPts[0].fX));
				rc.top = max(rcDirty.top, LONG(pPts[0].fY));
				rc.right = min(rcDirty.right, LONG(pPts[0].fX)+m_szSelSize.cx);
				rc.bottom = min(rcDirty.bottom, LONG(pPts[0].fY)+m_szSelSize.cy);

				for (LONG y = rc.top; y < rc.bottom; ++y)
				{
					TRasterImagePixel* pBuf = a_pBuffer+(y-rcDirty.top)*a_nStride+rc.left-rcDirty.left;
					TRasterImagePixel* pSel = m_cSelData.m_p+(y-LONG(pPts[0].fY))*m_szSelSize.cx+rc.left-LONG(pPts[0].fX);
					if (m_bUseDeleteMask)
					{
						BYTE* pDel = m_cDeleteMask.m_p+(y-LONG(pPts[0].fY))*(m_rcDelete.right-m_rcDelete.left)+rc.left-LONG(pPts[0].fX);
						switch (m_eBlendingMode)
						{
						case EBMDrawOver:
							for (LONG x = rc.left; x < rc.right; ++x, ++pBuf, ++pSel, ++pDel)
								CPixelMixerPaintOver::Mix(*pBuf, *pSel, *pDel);
							break;
						case EBMReplace:
							for (LONG x = rc.left; x < rc.right; ++x, ++pBuf, ++pSel, ++pDel)
								CPixelMixerReplace::Mix(*pBuf, *pSel, *pDel);
							break;
						case EBMDrawUnder:
							for (LONG x = rc.left; x < rc.right; ++x, ++pBuf, ++pSel, ++pDel)
								CPixelMixerPaintUnder::Mix(*pBuf, *pSel, *pDel);
							break;
						}
					}
					else
					{
						switch (m_eBlendingMode)
						{
						case EBMDrawOver:
							for (LONG x = rc.left; x < rc.right; ++x, ++pBuf, ++pSel)
								CPixelMixerPaintOver::Mix(*pBuf, *pSel);
							break;
						case EBMReplace:
							for (LONG x = rc.left; x < rc.right; ++x, ++pBuf, ++pSel)
								CPixelMixerReplace::Mix(*pBuf, *pSel);
							break;
						case EBMDrawUnder:
							for (LONG x = rc.left; x < rc.right; ++x, ++pBuf, ++pSel)
								CPixelMixerPaintUnder::Mix(*pBuf, *pSel);
							break;
						}
					}
				}
			}
			else
			{
				RECT rc = GetBoundingBox();
				bool const bSrcSmall = m_szSelSize.cx < 2 || m_szSelSize.cy < 2;
				if (rc.left < rcDirty.left) rc.left = rcDirty.left;
				if (rc.top < rcDirty.top) rc.top = rcDirty.top;
				if (rc.right > rcDirty.right) rc.right = rcDirty.right;
				if (rc.bottom > rcDirty.bottom) rc.bottom = rcDirty.bottom;
				double dest[8] = {pPts[0].fX, pPts[0].fY, pPts[1].fX, pPts[1].fY, pPts[2].fX, pPts[2].fY, pPts[3].fX, pPts[3].fY};
				TPixelCoords tCenter;
				float fSizeX;
				float fSizeY;
				float fAngle;
				if (m_cData.eMode == CEditToolDataTransformation::EMPerspective ||
					!GetRectangle(&tCenter, &fSizeX, &fSizeY, &fAngle))
				{
					agg::trans_perspective tr(dest, 0, 0, m_szSelSize.cx, m_szSelSize.cy);
					if (tr.is_valid())
						switch (m_eBlendingMode)
						{
						case EBMDrawOver:
							if (m_eRasterizationMode == ERMBinary || bSrcSmall)
								DrawSimple<CPixelMixerPaintOver>(rc, rcDirty, a_pBuffer, a_nStride, tr);
							else
								DrawSmooth<CPixelMixerPaintOver>(rc, rcDirty, a_pBuffer, a_nStride, tr);
							break;
						case EBMReplace:
							if (m_eRasterizationMode == ERMBinary || bSrcSmall)
								DrawSimple<CPixelMixerReplace>(rc, rcDirty, a_pBuffer, a_nStride, tr);
							else
								DrawSmooth<CPixelMixerReplace>(rc, rcDirty, a_pBuffer, a_nStride, tr);
							break;
						case EBMDrawUnder:
							if (m_eRasterizationMode == ERMBinary || bSrcSmall)
								DrawSimple<CPixelMixerPaintUnder>(rc, rcDirty, a_pBuffer, a_nStride, tr);
							else
								DrawSmooth<CPixelMixerPaintUnder>(rc, rcDirty, a_pBuffer, a_nStride, tr);
							break;
						}
				}
				else
				{
					agg::trans_affine tr;
					tr = agg::trans_affine_translation(-tCenter.fX, -tCenter.fY);
					tr *= agg::trans_affine_rotation(-fAngle/* * 3.1415926 / 180.0*/);
					tr *= agg::trans_affine_scaling(m_szSelSize.cx/fSizeX*0.5, m_szSelSize.cy/fSizeY*0.5);
					tr *= agg::trans_affine_translation(m_szSelSize.cx*0.5, m_szSelSize.cy*0.5);
					//agg::trans_bilinear tr(dest, 0, 0, m_szSelSize.cx, m_szSelSize.cy);
					if (tr.is_valid())
						switch (m_eBlendingMode)
						{
						case EBMDrawOver:
							if (m_eRasterizationMode == ERMBinary || bSrcSmall)
								DrawSimple<CPixelMixerPaintOver>(rc, rcDirty, a_pBuffer, a_nStride, tr);
							else
								DrawSmooth<CPixelMixerPaintOver>(rc, rcDirty, a_pBuffer, a_nStride, tr);
							break;
						case EBMReplace:
							if (m_eRasterizationMode == ERMBinary || bSrcSmall)
								DrawSimple<CPixelMixerReplace>(rc, rcDirty, a_pBuffer, a_nStride, tr);
							else
								DrawSmooth<CPixelMixerReplace>(rc, rcDirty, a_pBuffer, a_nStride, tr);
							break;
						case EBMDrawUnder:
							if (m_eRasterizationMode == ERMBinary || bSrcSmall)
								DrawSimple<CPixelMixerPaintUnder>(rc, rcDirty, a_pBuffer, a_nStride, tr);
							else
								DrawSmooth<CPixelMixerPaintUnder>(rc, rcDirty, a_pBuffer, a_nStride, tr);
							break;
						}
				}
			}
		}
		return S_OK;
	}
	HRESULT _IsDirty(RECT* a_pImageRect, BOOL* a_pOptimizeImageRect, RECT* a_pSelectionRect)
	{
		if (a_pOptimizeImageRect)
			*a_pOptimizeImageRect = TRUE; // TODO: compare number of pixels in rectangle with number of pixels in scanline buffers
		if (a_pSelectionRect)
		{
			a_pSelectionRect->left = 0;
			a_pSelectionRect->right = m_nSizeX;
			a_pSelectionRect->top = 0;
			a_pSelectionRect->bottom = m_nSizeY;
		}
		if (a_pImageRect)
		{
			*a_pImageRect = GetBoundingBox();
			if (a_pImageRect->left > m_rcDelete.left) a_pImageRect->left = m_rcDelete.left;
			if (a_pImageRect->top > m_rcDelete.top) a_pImageRect->top = m_rcDelete.top;
			if (a_pImageRect->right < m_rcDelete.right) a_pImageRect->right = m_rcDelete.right;
			if (a_pImageRect->bottom < m_rcDelete.bottom) a_pImageRect->bottom = m_rcDelete.bottom;
		}
		return m_cSelData.m_p ? S_OK : S_FALSE;
	}
	HRESULT _GetSelectionInfo(RECT* a_pBoundingRectangle, BOOL* a_bEntireRectangle)
	{
		if (a_pBoundingRectangle)
		{
			a_pBoundingRectangle->left = 0;
			a_pBoundingRectangle->right = m_nSizeX;
			a_pBoundingRectangle->top = 0;
			a_pBoundingRectangle->bottom = m_nSizeY;
		}
		if (a_bEntireRectangle)
		{
			*a_bEntireRectangle = TRUE;
		}
		return S_OK;
	}
	HRESULT _GetSelectionTile(LONG a_nX, LONG a_nY, ULONG a_nSizeX, ULONG a_nSizeY, ULONG a_nStride, BYTE* a_pBuffer)
	{
		for (ULONG y = 0; y < a_nSizeY; ++y)
			FillMemory(a_pBuffer+y*a_nStride, a_nSizeX, 0xff);
		return S_OK;
	}

	HRESULT _PreTranslateMessage(MSG const* a_pMsg)
	{
		if (S_OK == CEditToolRectangularShape<CEditToolTransformation>::_PreTranslateMessage(a_pMsg))
			return S_OK;

		if (a_pMsg->message != WM_KEYDOWN || a_pMsg->wParam != VK_DELETE || !RectangleDefined())
			return S_FALSE;

		// delete the selected region and apply
		m_cSelData.Free();
		return ETPAApply;
	}

	STDMETHOD(SetState)(ISharedState* a_pState)
	{
		CComQIPtr<CEditToolDataTransformation::ISharedStateToolData> pData(a_pState);
		if (pData)
		{
			m_cData = *(pData->InternalData());
			if (m_cData.bAspectLock && m_cData.eMode == CEditToolDataTransformation::EMSimple && m_szSelSize.cx && m_szSelSize.cy)
				SetAspectLock(m_szSelSize.cy/float(m_szSelSize.cx));
			bool bChange = SetRectangleMode(m_cData.eMode == CEditToolDataTransformation::EMSimple ? (m_cData.bAspectLock ? CEditToolRectangularShape::ERMRotatedLockAspectRectangle : CEditToolRectangularShape::ERMRotatedRectangle) : CEditToolRectangularShape::ERMCustom);
			if (bChange)
				ShapeChanged(false);
		}
		return S_OK;
	}
	HRESULT _SetGlobals(EBlendingMode a_eBlendingMode, ERasterizationMode a_eRasterizationMode, ECoordinatesMode a_eCoordinatesMode)
	{
		bool bBlendingChange = m_eBlendingMode != a_eBlendingMode;
		bool bRasterizationChange = m_eRasterizationMode != a_eRasterizationMode;
		bool bCoordinatesChange = m_eCoordinatesMode != a_eCoordinatesMode;

		if (!bBlendingChange && !bRasterizationChange && !bCoordinatesChange)
			return S_FALSE;
		m_eBlendingMode = a_eBlendingMode;
		m_eRasterizationMode = a_eRasterizationMode;
		m_eCoordinatesMode = a_eCoordinatesMode;
		if (!RectangleDefined())
			return S_OK;
		bool bRedrawCache = bRasterizationChange;
		if (bCoordinatesChange && AdjustRectangleCoordinates())
		{
			for (ULONG i = 0; i < 9; ++i)
				M_Window()->ControlPointChanged(i);
			M_Window()->ControlLinesChanged();
			bRedrawCache = true;
		}
		if (bRedrawCache)
		{
			ShapeChanged(false);
		}
		else if (bBlendingChange)
		{
			M_Window()->RectangleChanged(&m_rcPrev);
		}
		return S_OK;
	}

	void ShapeChanged(bool a_bInitFinish)
	{
		if (a_bInitFinish && m_cSelData.m_p == NULL)
		{
			m_cSelData.Free();
			m_cDeleteMask.Free();
			m_bUseDeleteMask = false;
			m_szSelSize.cx = m_szSelSize.cy = 0;
			RECT rcBounds = GetBoundingBox();
			if (rcBounds.left < 0) rcBounds.left = 0;
			if (rcBounds.top < 0) rcBounds.top = 0;
			if (rcBounds.right > LONG(m_nSizeX)) rcBounds.right = m_nSizeX;
			if (rcBounds.bottom > LONG(m_nSizeY)) rcBounds.bottom = m_nSizeY;
			if (rcBounds.left < rcBounds.right && rcBounds.top < rcBounds.bottom)
			{
				m_rcDelete = rcBounds;
				SetRectangle((rcBounds.left+rcBounds.right)*0.5f, (rcBounds.top+rcBounds.bottom)*0.5f, (rcBounds.right-rcBounds.left)*0.5f, (rcBounds.bottom-rcBounds.top)*0.5f, 0.0f);
				m_rcPrev = GetBoundingBox();
				m_szSelSize.cx = rcBounds.right-rcBounds.left;
				m_szSelSize.cy = rcBounds.bottom-rcBounds.top;
				SetAspectLock(m_szSelSize.cy/float(m_szSelSize.cx));
				m_cSelData.Allocate(m_szSelSize.cx*m_szSelSize.cy);
				M_Window()->GetImageTile(rcBounds.left, rcBounds.top, rcBounds.right-rcBounds.left, rcBounds.bottom-rcBounds.top, 0.0f, rcBounds.right-rcBounds.left, EITIContent, m_cSelData);
				M_Window()->RectangleChanged(&rcBounds);
				M_Window()->ControlPointsChanged();
				M_Window()->ControlLinesChanged();
			}
			else
			{
				ResetRectangle();
			}
		}
		else
		{
			RECT rcPrev = m_rcPrev; //GetBoundingBox()
			RECT rc = GetBoundingBox();
			m_rcPrev = rc;
			if (rc.left > rcPrev.left) rc.left = rcPrev.left;
			if (rc.top > rcPrev.top) rc.top = rcPrev.top;
			if (rc.right < rcPrev.right) rc.right = rcPrev.right;
			if (rc.bottom < rcPrev.bottom) rc.bottom = rcPrev.bottom;
			M_Window()->RectangleChanged(&rc);
		}
	}

	HRESULT _AdjustCoordinates(EControlKeysState UNREF(a_eKeysState), TPixelCoords* a_pPos, TPixelCoords const* a_pPointerSize, ULONG const* a_pControlPointIndex)
	{
		if (m_eCoordinatesMode == ECMIntegral)
		{
			a_pPos->fX = floorf(a_pPos->fX+0.5f);
			a_pPos->fY = floorf(a_pPos->fY+0.5f);
		}
		return S_OK;
	}

	bool HitTest(float a_fX, float a_fY) const
	{
		return RectangleHitTest(a_fX, a_fY);
	}
	bool UseMoveCursor(TPixelCoords const* a_pPos) const
	{
		return RectangleMoving() || (!M_Dragging() && HitTest(a_pPos->fX, a_pPos->fY));
	}
	STDMETHOD(PointTest)(EControlKeysState UNREF(a_eKeysState), TPixelCoords const* a_pPos, BYTE UNREF(a_bAccurate))
	{
		return HitTest(a_pPos->fX, a_pPos->fY) ? ETPAHit|ETPATransform : ETPAMissed|ETPAStartNew;
	}
	STDMETHOD(Transform)(TMatrix3x3f const* a_pMatrix)
	{
		return E_NOTIMPL;
	}
	void DeleteCachedData() {}

	// IRasterImageEditToolScripting
public:
	void InitScriptSel(int a_nX1 = 0, int a_nY1 = 0, int a_nX2 = 0x7ffffff, int a_nY2 = 0x7ffffff)
	{
		if (a_nX1 > a_nX2)
		{
			int n = a_nX1;
			a_nX1 = a_nX2;
			a_nX2 = n;
		}
		if (a_nY1 > a_nY2)
		{
			int n = a_nY1;
			a_nY1 = a_nY2;
			a_nY2 = n;
		}
		if (a_nX1 < 0) a_nX1 = 0;
		if (a_nY1 < 0) a_nY1 = 0;
		ULONG nSizeX = 0;
		ULONG nSizeY = 0;
		M_Window()->Size(&nSizeX, &nSizeY);
		if (a_nX2 > int(nSizeX)) a_nX2 = nSizeX;
		if (a_nY2 > int(nSizeY)) a_nY2 = nSizeY;
		m_cSelData.Free();
		m_cDeleteMask.Free();
		m_bUseDeleteMask = false;
		m_szSelSize.cx = m_szSelSize.cy = 0;
		m_rcDelete.left = a_nX1;
		m_rcDelete.top = a_nY1;
		m_rcDelete.right = a_nX2;
		m_rcDelete.bottom = a_nY2;
		m_szSelSize.cx = m_rcDelete.right-m_rcDelete.left;
		m_szSelSize.cy = m_rcDelete.bottom-m_rcDelete.top;
		m_cSelData.Allocate(m_szSelSize.cx*m_szSelSize.cy);
		M_Window()->GetImageTile(a_nX1, a_nY1, a_nX2-a_nX1, a_nY2-a_nY1, 0.0f, a_nX2-a_nX1, EITIContent, m_cSelData);
	}
	STDMETHOD(FromText)(BSTR a_bstrParams)
	{
		try
		{
			if (a_bstrParams == NULL)
				return S_FALSE;
			float aVals[12];
			switch (swscanf(a_bstrParams, L"%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f", aVals, aVals+1, aVals+2, aVals+3, aVals+4, aVals+5, aVals+6, aVals+7, aVals+8, aVals+9, aVals+10, aVals+11))
			{
			case 2: // whole: move only
				InitScriptSel();
				m_cData.eMode = CEditToolDataTransformation::EMSimple;
				{
					ULONG nSizeX = 0;
					ULONG nSizeY = 0;
					M_Window()->Size(&nSizeX, &nSizeY);
					SetRectangle(aVals[0]+nSizeX*0.5f, aVals[1]+nSizeY*0.5f, nSizeX*0.5f, nSizeY*0.5f, 0.0f);
				}
				break;
			case 5: // whole: move, scale, rotate
				InitScriptSel();
				m_cData.eMode = CEditToolDataTransformation::EMSimple;
				{
					ULONG nSizeX = 0;
					ULONG nSizeY = 0;
					M_Window()->Size(&nSizeX, &nSizeY);
					SetRectangle(aVals[0]+nSizeX*0.5f, aVals[1]+nSizeY*0.5f, aVals[2]*0.5f, aVals[3]*0.5f, aVals[4]*3.141593f/180.0f);
				}
				break;
			case 8: // whole: perspective
				InitScriptSel();
				m_cData.eMode = CEditToolDataTransformation::EMPerspective;
				SetRectanglePoints(aVals[0], aVals[1], aVals[2], aVals[3], aVals[4], aVals[5], aVals[6], aVals[7]);
				break;
			case 6: // rect: move only
				InitScriptSel(aVals[0]+0.5f, aVals[1]+0.5f, aVals[2]+0.5f, aVals[3]+0.5f);
				m_cData.eMode = CEditToolDataTransformation::EMSimple;
				{
					ULONG nSizeX = 0;
					ULONG nSizeY = 0;
					M_Window()->Size(&nSizeX, &nSizeY);
					SetRectangle(aVals[4]+nSizeX*0.5f, aVals[5]+nSizeY*0.5f, abs(int(aVals[2]+0.5f)-int(aVals[0]+0.5f))*0.5f, abs(int(aVals[3]+0.5f)-int(aVals[1]+0.5f))*0.5f, 0.0f);
				}
				break;
			case 9: // rect: move, scale, rotate
				InitScriptSel(aVals[0]+0.5f, aVals[1]+0.5f, aVals[2]+0.5f, aVals[3]+0.5f);
				m_cData.eMode = CEditToolDataTransformation::EMSimple;
				{
					ULONG nSizeX = 0;
					ULONG nSizeY = 0;
					M_Window()->Size(&nSizeX, &nSizeY);
					SetRectangle(aVals[4]+nSizeX*0.5f, aVals[5]+nSizeY*0.5f, aVals[6]*0.5f, aVals[7]*0.5f, aVals[8]*3.141593f/180.0f);
				}
				break;
			case 12: // rect: perspective
				InitScriptSel(aVals[0]+0.5f, aVals[1]+0.5f, aVals[2]+0.5f, aVals[3]+0.5f);
				m_cData.eMode = CEditToolDataTransformation::EMPerspective;
				SetRectanglePoints(aVals[4], aVals[5], aVals[6], aVals[7], aVals[8], aVals[9], aVals[10], aVals[11]);
				break;
			default:
				return E_FAIL;
			}
			return S_OK;
		}
		catch (...)
		{
			return E_UNEXPECTED;
		}
	}
	STDMETHOD(ToText)(BSTR* a_pbstrParams)
	{
		return E_NOTIMPL;
	}

	// IRasterImageEditToolFloatingSelection methods
public:
	STDMETHOD(Set)(float a_fX, float a_fY, ULONG a_nSizeX, ULONG a_nSizeY, ULONG a_nStride, TRasterImagePixel const* a_pBuffer, BOOL a_bModifyExisting)
	{
		//if (_IsDirty(NULL, NULL, NULL) == S_OK)
		//	M_Window()->ApplyChanges();
		if (a_bModifyExisting && a_nSizeX == m_szSelSize.cx && a_nSizeX == m_szSelSize.cx)
		{
			for (ULONG y = 0; y < a_nSizeY; ++y)
				CopyMemory(m_cSelData.m_p+y*a_nSizeX, a_pBuffer+y*a_nStride, a_nSizeX*sizeof *a_pBuffer);
			M_Window()->RectangleChanged(&m_rcPrev);
			return S_OK;
		}

		m_cSelData.Free();
		m_cSelMask.Free();
		m_bUseDeleteMask = false;
		if (!a_bModifyExisting)
		{
			m_cDeleteMask.Free();
			m_rcDelete = RECT_EMPTY;
		}
		m_szSelSize.cx = a_nSizeX;
		m_szSelSize.cy = a_nSizeY;
		if (m_cData.bAspectLock && m_cData.eMode == CEditToolDataTransformation::EMSimple && m_szSelSize.cx && m_szSelSize.cy)
			SetAspectLock(m_szSelSize.cy/float(m_szSelSize.cx));
		if (a_nSizeX && a_nSizeY && a_pBuffer)
		{
			m_cSelData.Allocate(a_nSizeX*a_nSizeY);
			for (ULONG y = 0; y < a_nSizeY; ++y)
				CopyMemory(m_cSelData.m_p+y*a_nSizeX, a_pBuffer+y*a_nStride, a_nSizeX*sizeof *a_pBuffer);
		}
		if (a_bModifyExisting)
		{
			if (m_cData.eMode == CEditToolDataTransformation::EMSimple)
			{
				TPixelCoords tC;
				float fX, fY, fA;
				GetRectangle(&tC, &fX, &fY, &fA);
				SetRectangle(a_fX-fX+a_nSizeX*0.5f, a_fY-fY+a_nSizeY*0.5f, a_nSizeX*0.5f, a_nSizeY*0.5f, fA);
			}
			else if (m_cData.eMode == CEditToolDataTransformation::EMPerspective)
			{
				TPixelCoords const* pPts = GetRectanglePoints();
				float fX = 0.25*(pPts[0].fX+pPts[1].fX+pPts[2].fX+pPts[3].fX)-a_fX;
				float fY = 0.25*(pPts[0].fY+pPts[1].fY+pPts[2].fY+pPts[3].fY)-a_fY;
				SetRectanglePoints(pPts[0].fX+fX, pPts[0].fY+fY, pPts[1].fX+fX, pPts[1].fY+fY, pPts[2].fX+fX, pPts[2].fY+fY, pPts[3].fX+fX, pPts[3].fY+fY);
			}
		}
		else
		{
			SetRectangle(a_fX+a_nSizeX*0.5f, a_fY+a_nSizeY*0.5f, a_nSizeX*0.5f, a_nSizeY*0.5f, 0.0f);
		}
		return S_OK;
	}
	STDMETHOD(Position)(float* a_pX, float* a_pY)
	{
		if (m_cData.eMode == CEditToolDataTransformation::EMSimple)
		{
			TPixelCoords tC;
			float fX, fY, fA;
			GetRectangle(&tC, &fX, &fY, &fA);
			*a_pX = tC.fX;
			*a_pY = tC.fY;
			return S_OK;
		}
		else if (m_cData.eMode == CEditToolDataTransformation::EMPerspective)
		{
			TPixelCoords const* pPts = GetRectanglePoints();
			*a_pX = 0.25*(pPts[0].fX+pPts[1].fX+pPts[2].fX+pPts[3].fX);
			*a_pY = 0.25*(pPts[0].fY+pPts[1].fY+pPts[2].fY+pPts[3].fY);
			return S_OK;
		}
		return E_NOTIMPL;
	}
	STDMETHOD(Size)(ULONG* a_pSizeX, ULONG* a_pSizeY)
	{
		*a_pSizeX = m_szSelSize.cx;
		*a_pSizeY = m_szSelSize.cy;
		return S_OK;
	}
	STDMETHOD(Data)(ULONG a_nX, ULONG a_nY, ULONG a_nSizeX, ULONG a_nSizeY, ULONG a_nStride, TRasterImagePixel* a_pBuffer)
	{
		try
		{
			if (LONG(a_nX+a_nSizeX) > m_szSelSize.cx || LONG(a_nY+a_nSizeY) > m_szSelSize.cy)
				return E_RW_INVALIDPARAM;
			if (m_cDeleteMask.m_p)
			{
				for (ULONG y = 0; y < a_nSizeY; ++y)
				{
					TRasterImagePixel* pD = a_pBuffer+y*a_nStride;
					TRasterImagePixel const* pS = m_cSelData.m_p+(y+a_nY)*m_szSelSize.cx+a_nX;
					BYTE const* pM = m_cDeleteMask.m_p+(y+a_nY)*m_szSelSize.cx+a_nX;
					for (ULONG x = 0; x < a_nSizeX; ++x)
					{
						pD->bA = pS->bA*ULONG(*pM)/255;
						if (pD->bA)
						{
							pD->bR = pS->bR;
							pD->bG = pS->bG;
							pD->bB = pS->bB;
						}
						else
						{
							pD->bR = pD->bG = pD->bB = 0;
						}
						++pD;
						++pS;
						++pM;
					}
				}
			}
			else
			{
				for (ULONG y = 0; y < a_nSizeY; ++y)
					CopyMemory(a_pBuffer+y*a_nStride, m_cSelData.m_p+(y+a_nY)*m_szSelSize.cx+a_nX, a_nSizeX*sizeof *m_cSelData.m_p);
			}
			return S_OK;
		}
		catch (...)
		{
			return a_pBuffer ? E_UNEXPECTED : E_POINTER;
		}
	}
	STDMETHOD(Delete)()
	{
		m_szSelSize.cx = m_szSelSize.cy = 0;
		m_cSelData.Free();
		m_cSelMask.Free();
		m_bUseDeleteMask = false;
		return ETPAApply;
	}
	STDMETHOD(SelectAll)()
	{
		m_cSelData.Free();
		m_cDeleteMask.Free();
		m_bUseDeleteMask = false;
		m_rcDelete.left = 0;
		m_rcDelete.top = 0;
		m_rcDelete.right = m_nSizeX;
		m_rcDelete.bottom = m_nSizeY;
		SetRectangle(m_nSizeX*0.5f, m_nSizeY*0.5f, m_nSizeX*0.5f, m_nSizeY*0.5f, 0.0f);
		m_rcPrev = GetBoundingBox();
		m_szSelSize.cx = m_nSizeX;
		m_szSelSize.cy = m_nSizeY;
		m_cSelData.Allocate(m_szSelSize.cx*m_szSelSize.cy);
		M_Window()->GetImageTile(0, 0, m_nSizeX, m_nSizeY, 0.0f, m_nSizeX, EITIContent, m_cSelData);
		RECT rcAll = {0, 0, m_nSizeX, m_nSizeY};
		if (m_cData.bAspectLock && m_cData.eMode == CEditToolDataTransformation::EMSimple && m_szSelSize.cx && m_szSelSize.cy)
			SetAspectLock(m_szSelSize.cy/float(m_szSelSize.cx));
		M_Window()->RectangleChanged(&rcAll);
		M_Window()->ControlPointsChanged();
		M_Window()->ControlLinesChanged();
		return S_OK;
	}

	// CEditToolRectangularShape overrides
public:
	TPixelCoords const* CRGetRectanglePoints() const { return m_aPerspective; }
	void CRCustomToRectangle(TPixelCoords* a_pCenter, float* a_pSizeX, float* a_pSizeY, float* a_pAngle, bool a_bSquare) const
	{
		RECT rc = GetBoundingBox();
		a_pCenter->fX = 0.5f*(rc.left+rc.right);
		a_pCenter->fY = 0.5f*(rc.top+rc.bottom);
		*a_pAngle = 0.0f;
		if (a_bSquare)
		{
			float f = sqrtf((rc.right-rc.left)*(rc.right-rc.left) + (rc.bottom-rc.top)*(rc.bottom-rc.top));
			*a_pSizeX = f/sqrtf(GetAspectLock());
			*a_pSizeY = f*sqrtf(GetAspectLock());
			if (M_CoordinatesMode() == ECMIntegral)
			{
				*a_pSizeX = floorf(*a_pSizeX+0.5f);
				*a_pSizeY = floorf(*a_pSizeY+0.5f);
			}
		}
		else
		{
			*a_pSizeX = rc.right-rc.left;
			*a_pSizeY = rc.bottom-rc.top;
		}
	}
	void CRMoveRectangle(float a_fDX, float a_fDY)
	{
		SetRectanglePoints(m_aPerspective[0].fX+a_fDX, m_aPerspective[0].fY+a_fDY, m_aPerspective[1].fX+a_fDX, m_aPerspective[1].fY+a_fDY, m_aPerspective[2].fX+a_fDX, m_aPerspective[2].fY+a_fDY, m_aPerspective[3].fX+a_fDX, m_aPerspective[3].fY+a_fDY);
	}
	HRESULT CRGetControlLines(IEditToolControlLines* a_pLines, ULONG a_nLineTypes) const
	{
		if (a_nLineTypes&ECLTHelp)
		{
			a_pLines->MoveTo(m_aPerspective[0].fX, m_aPerspective[0].fY);
			a_pLines->LineTo(m_aPerspective[1].fX, m_aPerspective[1].fY);
			a_pLines->MoveTo(m_aPerspective[0].fX, m_aPerspective[0].fY);
			a_pLines->LineTo(m_aPerspective[3].fX, m_aPerspective[3].fY);
			a_pLines->MoveTo(m_aPerspective[2].fX, m_aPerspective[2].fY);
			a_pLines->LineTo(m_aPerspective[1].fX, m_aPerspective[1].fY);
			a_pLines->MoveTo(m_aPerspective[2].fX, m_aPerspective[2].fY);
			a_pLines->LineTo(m_aPerspective[3].fX, m_aPerspective[3].fY);
		}
		return S_OK;
	}
	HRESULT CRSetControlPoint(ULONG a_nIndex, TPixelCoords const* a_pPos, boolean a_bFinished, float a_fPointSize)
	{
		if (a_nIndex&1)
		{
			ULONG nPt1 = a_nIndex>>1;
			ULONG nPt2 = ((a_nIndex+1)>>1)&3;
			float fPX = 0.5f*(m_aPerspective[nPt1].fX+m_aPerspective[nPt2].fX);
			float fPY = 0.5f*(m_aPerspective[nPt1].fY+m_aPerspective[nPt2].fY);
			m_aPerspective[nPt1].fX += a_pPos->fX-fPX;
			m_aPerspective[nPt2].fX += a_pPos->fX-fPX;
			m_aPerspective[nPt1].fY += a_pPos->fY-fPY;
			m_aPerspective[nPt2].fY += a_pPos->fY-fPY;
		}
		else
		{
			m_aPerspective[a_nIndex>>1] = *a_pPos;
		}
		return S_OK;
	}
	HRESULT CRGetControlPoint(ULONG a_nIndex, TPixelCoords* a_pPos, ULONG* a_pClass) const
	{
		return S_OK;
	}
	HRESULT CRGetControlPointCount(ULONG* a_pCount) const
	{
		*a_pCount = 4;
		return S_OK;
	}
	void SetRectanglePoints(float a_fX1, float a_fY1, float a_fX2, float a_fY2, float a_fX3, float a_fY3, float a_fX4, float a_fY4)
	{
		SetRectangleMode(ERMCustom);
		m_aPerspective[0].fX = a_fX1;
		m_aPerspective[0].fY = a_fY1;
		m_aPerspective[1].fX = a_fX2;
		m_aPerspective[1].fY = a_fY2;
		m_aPerspective[2].fX = a_fX3;
		m_aPerspective[2].fY = a_fY3;
		m_aPerspective[3].fX = a_fX4;
		m_aPerspective[3].fY = a_fY4;
		CustomRectangleChanged();
	}

private:
	EBlendingMode m_eBlendingMode;
	ERasterizationMode m_eRasterizationMode;
	ECoordinatesMode m_eCoordinatesMode;
	CEditToolDataTransformation m_cData;
	ULONG m_nSizeX;
	ULONG m_nSizeY;
	RECT m_rcPrev;

	// region to delete
	RECT m_rcDelete;
	CAutoVectorPtr<BYTE> m_cDeleteMask;

	// floating selection
	SIZE m_szSelSize;
	CAutoVectorPtr<TRasterImagePixel> m_cSelData;
	CAutoVectorPtr<BYTE> m_cSelMask;
	bool m_bUseDeleteMask;

	TPixelCoords m_aPerspective[4];
	TPixelCoords m_aBezier[16];
};

