
#pragma once

struct ImageTile;

struct RWHash
{
	ULONGLONG val[4];
};

typedef ImageTile const* (fnUpdateTile*)(void* context, LONG tx, LONG ty, ImageTile const* old);

template<typename T>
struct RWObjPtr
{
private:
	T* p;
};

MIDL_INTERFACE("55630B53-B693-493A-98E7-0437FB4B620F")
IImageChannel32f2 : public IUnknown
{
public:
	virtual RWHash const& STDMETHODCALLTYPE Hash();
	virtual EImageChannelID STDMETHODCALLTYPE ChannelID();
	virtual RWObjPtr<ImageTile const> STDMETHODCALLTYPE Background();
	virtual RECT STDMETHODCALLTYPE Tiles();
	virtual RWObjPtr<ImageTile const>* inner;

	virtual ImageChannel32f* STDMETHODCALLTYPE Copy() const;
	virtual void STDMETHODCALLTYPE Finalize(bool updateHash);

	virtual HRESULT STDMETHODCALLTYPE DefaultSet(ImageTile const* tile, ImageChannel32f const** res) const;
	virtual HRESULT STDMETHODCALLTYPE TilesReplace(RECT rect, ImageTile const* tiles, ImageChannel32f const** res) const;
	virtual HRESULT STDMETHODCALLTYPE TilesUpdate(RECT rect, fnUpdateTile updater, ImageChannel32f const** res) const;

	virtual HRESULT STDMETHODCALLTYPE DefaultSet(ImageTile const* tile);
	virtual HRESULT STDMETHODCALLTYPE TilesReplace(RECT rect, ImageTile const* tiles);
	virtual HRESULT STDMETHODCALLTYPE TilesUpdate(RECT rect, fnUpdateTile updater);
};

MIDL_INTERFACE("55630B53-B693-493A-98E7-0437FB4B620F")
IImageChannel32f : public IUnknown
{
public:
	RWHash hash;
	EImageChannelID channelID;
	RWObjPtr<ImageTile const> outer; // default
	RECT tiles; // inner tiles in grid coords (size of 2x2 means 128x128 pixels with 64x64 tiles)
	RWObjPtr<ImageTile const>* inner; // array of size tiles.width * tiles.height

	virtual ImageChannel32f* STDMETHODCALLTYPE Copy() const;
	virtual void STDMETHODCALLTYPE Finalize(bool updateHash);

	virtual HRESULT STDMETHODCALLTYPE DefaultSet(ImageTile const* tile, ImageChannel32f const** res) const;
	virtual HRESULT STDMETHODCALLTYPE TilesReplace(RECT rect, ImageTile const* tiles, ImageChannel32f const** res) const;
	virtual HRESULT STDMETHODCALLTYPE TilesUpdate(RECT rect, fnUpdateTile updater, ImageChannel32f const** res) const;

	virtual HRESULT STDMETHODCALLTYPE DefaultSet(ImageTile const* tile);
	virtual HRESULT STDMETHODCALLTYPE TilesReplace(RECT rect, ImageTile const* tiles);
	virtual HRESULT STDMETHODCALLTYPE TilesUpdate(RECT rect, fnUpdateTile updater);
};

MIDL_INTERFACE("F91FF803-025F-4878-9688-69D9B2850F5E")
IImageChannelEditable32f : public IUnknown
{
public:
	RWHash hash;
	EImageChannelID channelID;
	ImageTile const* outer; // default 
	RECT tiles; // inner tiles in grid coords (size of 2x2 means 128x128 pixels with 64x64 tiles)
	ImageTile const* inner; // array of size tiles.width * tiles.height

	virtual ImageChannel32f* STDMETHODCALLTYPE Copy() const;
	virtual void STDMETHODCALLTYPE Finalize(bool updateHash);

	virtual HRESULT STDMETHODCALLTYPE DefaultSet(ImageTile const* tile);
	virtual HRESULT STDMETHODCALLTYPE TilesReplace(RECT rect, ImageTile const* tiles);
	virtual HRESULT STDMETHODCALLTYPE TilesUpdate(RECT rect, fnUpdateTile updater);
};

class CImageChannel32f : public IImageChannel32f
{
public:
	CImageChannel32f* Copy() const
	{
		CComObject<CImageChannel32f>* p = NULL;
		CComObject<CImageChannel32f>::CreateInstance(&p);
		CComPtr<IImageChannel32f> tmp(p);
		p->hash = hash;
		p->channelID = channelID;
		p->outer = outer;
		if (p->outer)
			p->outer->AddRef();
		p->tiles = tiles;
		size_t count = p->tiles->left < p->tiles->right && p->tiles->top < p->tiles->bottom ?
			(p->tiles->right-p->tiles->left)*(p->tiles->bottom-p->tiles->top) : 0;
		p->inner = count ? new ImageTile const[count] : NULL;
		for (size_t i = 0; i < count; ++i)
		{
			p->inner[i] = inner[i];
			if (p->inner[i])
				p->inner[i]->AddRef();
		}
		return tmp.Detach();
	}
	void Finalize(bool updateHash) const
	{
		if (updateHash)
		{
			RWHash h;
			h.Add(&channelID);
			if (outer)
				h.Add(outer->hash);
			h.Add(&rect);

		}
	}

	HRESULT DefaultSet(ImageTile const* tile)
	{
		if (outer)
			outer->fnRelease(outer);
		outer = tile;
		if (outer)
			outer->fnAddRef(outer);
		return S_OK;
	}
	HRESULT DefaultSet(ImageTile const* tile, ImageChannel32f const** res) const
	{
		CComPtr<ImageChannel32f> p;
		p.Attach(Copy());
		HRESULT ret = p->DefaultSet(tile);
		if (SUCCEEDED(ret))
			*res = p.Detach();
		return ret;
	}
};

MIDL_INTERFACE("258CCB70-8CCA-4246-B91D-B38C712AB205")
IRasterImage32f : public IUnknown
{
public:
	RWHash hash;
	ULONG channelCount;
	ImageChannel32f const* channels;

	virtual IRasterImage32f* STDMETHODCALLTYPE Copy() const;
	virtual IRasterImage32f* STDMETHODCALLTYPE ChannelRemove(ImageChannel channel);
	virtual HRESULT STDMETHODCALLTYPE ChannelRemove(ImageChannel channel, IRasterImage32f const** res) const;
	virtual HRESULT STDMETHODCALLTYPE ChannelsRemove(ULONG count, ImageChannel const* channels, IRasterImage32f const** res);
	virtual HRESULT STDMETHODCALLTYPE ChannelsAdd(ULONG count, ImageChannel const* channels, IRasterImage32f const** res);
};

IRasterImage32fEditablePtr IRasterImage32fCreate(ULONG channels)
{
}

MIDL_INTERFACE("DEBBDAA5-49FC-4CE5-BFA7-A5CED080A597")
IDocumentRasterImage32f : public IUnknown
{
public:
	virtual IRasterImage32f* STDMETHODCALLTYPE Get();
	virtual HRESULT STDMETHODCALLTYPE Set(IRasterImage32f* a_pImage);

	virtual HRESULT STDMETHODCALLTYPE ObserverIns(IRasterImage32fObserver* a_pObserver, TCookie a_tCookie);
	virtual HRESULT STDMETHODCALLTYPE ObserverDel(IRasterImage32fObserver* a_pObserver, TCookie a_tCookie);
};
