
void AddStringToVersion(BYTE*& a_pString, LPCOLESTR a_pszName, LPCOLESTR a_pszValue, int a_nValLen = -1)
{
	if (a_nValLen < 0)
		a_nValLen = wcslen(a_pszValue);
	*reinterpret_cast<WORD*>(a_pString+2) = a_nValLen;
	a_pString[4] = 1;
	wcscpy(reinterpret_cast<wchar_t*>(a_pString+6), a_pszName);
	ULONG nValOff = (wcslen(a_pszName)+5)&~1;
	wcsncpy(reinterpret_cast<wchar_t*>(a_pString)+nValOff, a_pszValue, a_nValLen);
	*reinterpret_cast<WORD*>(a_pString) = 2*(nValOff + a_nValLen+1);
	a_pString += 2*(nValOff + ((a_nValLen+2)&~1));
}

#pragma pack(push)
#pragma pack(2)
struct GRPICONDIRENTRY
{
   BYTE   bWidth;               // Width, in pixels, of the image
   BYTE   bHeight;              // Height, in pixels, of the image
   BYTE   bColorCount;          // Number of colors in image (0 if >=8bpp)
   BYTE   bReserved;            // Reserved
   WORD   wPlanes;              // Color Planes
   WORD   wBitCount;            // Bits per pixel
   DWORD  dwBytesInRes;         // how many bytes in this resource?
   WORD   nID;                  // the ID
};

struct GRPICONDIR
{
   WORD            idReserved;   // Reserved (must be 0)
   WORD            idType;       // Resource type (1 for icons)
   WORD            idCount;      // How many images?
   GRPICONDIRENTRY idEntries[3]; // The entries for each image
};
#pragma pack(pop)

void GetIconBits(HICON a_hIcon, BYTE* a_pIconBits, int a_nSize)
{
	ULONG const nXORSize = a_nSize*a_nSize*4;
	int const nANDLine = ((((a_nSize+7)>>3)+3)&0xfffffffc);
	ULONG const nANDSize = a_nSize*nANDLine;
	ZeroMemory(a_pIconBits, nXORSize+nANDSize+sizeof BITMAPINFOHEADER);
	ICONINFO tII;
	ZeroMemory(&tII, sizeof tII);
	GetIconInfo(a_hIcon, &tII);
	if (tII.hbmMask) DeleteObject(tII.hbmMask);
	HDC hDC = GetDC(NULL);
	BITMAPINFO tBMI;
	ZeroMemory(&tBMI, sizeof tBMI);
	tBMI.bmiHeader.biSize = sizeof tBMI.bmiHeader;
	tBMI.bmiHeader.biWidth = a_nSize;
	tBMI.bmiHeader.biHeight = a_nSize;
	tBMI.bmiHeader.biPlanes = 1;
	tBMI.bmiHeader.biBitCount = 32;
	tBMI.bmiHeader.biCompression = BI_RGB;
	GetDIBits(hDC, tII.hbmColor, 0, a_nSize, a_pIconBits+sizeof BITMAPINFOHEADER, &tBMI, DIB_RGB_COLORS);
	if (tII.hbmColor) DeleteObject(tII.hbmColor);
	ReleaseDC(NULL, hDC);
	BITMAPINFOHEADER* pHead = reinterpret_cast<BITMAPINFOHEADER*>(a_pIconBits);
	pHead->biSize = sizeof(BITMAPINFOHEADER);
	pHead->biWidth = a_nSize;
	pHead->biHeight = a_nSize<<1;
	pHead->biPlanes = 1;
	pHead->biBitCount = 32;
	pHead->biCompression = BI_RGB;
	pHead->biSizeImage = nANDSize+nXORSize;
	pHead->biXPelsPerMeter = 0;
	pHead->biYPelsPerMeter = 0;
	DWORD* pXOR = reinterpret_cast<DWORD*>(a_pIconBits+sizeof BITMAPINFOHEADER);
	BYTE *pAND = reinterpret_cast<BYTE*>(pXOR+(a_nSize*a_nSize));
	for (int y = 0; y < a_nSize; ++y)
	{
		for (int x = 0; x < a_nSize; ++x)
		{
			if (0 == pXOR&0xff000000)
			{
				pAND[x>>3] |= 0x80 >> (x&7);
			}
			++pXOR;
		}
		pAND += nANDLine;
	}
}

bool CreateDroplet(BYTE const* a_pOperData, ULONG a_nOperLen, LPCOLESTR a_pszDescription, ULONG a_nDescLen, HICON a_hIcSmall, HICON a_hIcMedium, HICON a_hIcLarge, IStorageFilter* a_pStorage, bool a_bOpenTarget, bool a_bWaitForEnter, bool a_bEnableGUI)
{
	TCHAR szTemp[MAX_PATH];
	{
		HRSRC hRsrc = FindResource(_pModule->get_m_hInst(), MAKEINTRESOURCE(IDR_COMMANDLINEDROPLET), _T("DROPLET"));
		if (hRsrc == NULL)
			return false;
		ULONG nSize = SizeofResource(_pModule->get_m_hInst(), hRsrc);
		void* pData = LockResource(LoadResource(_pModule->get_m_hInst(), hRsrc));
		TCHAR szTempPath[MAX_PATH];
		GetTempPath(MAX_PATH, szTempPath);
		GetTempFileName(szTempPath, _T("DRP"), 0, szTemp);
		HANDLE hDrp = CreateFile(szTemp, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, 0, NULL);
		DWORD dwWritten;
		WriteFile(hDrp, pData, nSize, &dwWritten, NULL);
		CloseHandle(hDrp);
	}
	// add operation to droplet template as a resource
	HANDLE hRes = BeginUpdateResource(szTemp, FALSE);
	if (NULL == hRes)
	{
		DeleteFile(szTemp);
		return false;
	}
	UpdateResource(hRes, _T("RWBATCHOP"), MAKEINTRESOURCE(1), 0, const_cast<BYTE*>(a_pOperData), a_nOperLen);
	BYTE aOptions[3] =
	{
		a_bEnableGUI ? 1 : 0,
		a_bWaitForEnter ? 1 : 0,
		a_bOpenTarget ? 1 : 0,
	};
	UpdateResource(hRes, _T("RWOPTIONS"), MAKEINTRESOURCE(1), 0, aOptions, 3);
	BYTE szVersionInfo[2048]; // should be enough
	ZeroMemory(szVersionInfo, sizeof szVersionInfo);
	szVersionInfo[2] = sizeof(VS_FIXEDFILEINFO);
	wcscpy(reinterpret_cast<wchar_t*>(szVersionInfo+6), L"VS_VERSION_INFO");
	VS_FIXEDFILEINFO* pFixedInfo = reinterpret_cast<VS_FIXEDFILEINFO*>(szVersionInfo+0x28);
	pFixedInfo->dwSignature = 0xfeef04bd;
	pFixedInfo->dwStrucVersion = 0x100;
	pFixedInfo->dwFileVersionMS = 0x40001;
	pFixedInfo->dwFileVersionLS = 0;
	pFixedInfo->dwProductVersionMS = 0x7d80001; // 2008.1
	pFixedInfo->dwProductVersionLS = 0;
	pFixedInfo->dwFileFlagsMask = VS_FFI_FILEFLAGSMASK;
#ifdef _DEBUG
	pFixedInfo->dwFileFlags = VS_FF_DEBUG;
#else
	pFixedInfo->dwFileFlags = 0;
#endif
	pFixedInfo->dwFileOS = VOS__WINDOWS32;
	pFixedInfo->dwFileType = VFT_UNKNOWN;
	pFixedInfo->dwFileSubtype = VFT2_UNKNOWN;
	pFixedInfo->dwFileDateMS = 0;
	pFixedInfo->dwFileDateLS = 0;
	BYTE* pStringInfo = szVersionInfo+0x5c;
	// string file info
	pStringInfo[4] = 1;
	wcscpy(reinterpret_cast<wchar_t*>(pStringInfo+6), L"StringFileInfo");
	BYTE* pStringTable = pStringInfo+0x24;
	pStringTable[4] = 1;
	wcscpy(reinterpret_cast<wchar_t*>(pStringTable+6), L"040904b0");
	BYTE* pString = pStringTable+0x18;
	AddStringToVersion(pString, L"FileDescription", a_pszDescription, a_nDescLen);
	AddStringToVersion(pString, L"ProductName", L"RealWorld Photos");
	AddStringToVersion(pString, L"ProductVersion", L"2008.1");
	*reinterpret_cast<WORD*>(pStringTable) = pString-pStringTable;
	*reinterpret_cast<WORD*>(pStringInfo) = pString-pStringInfo;
	BYTE* pVarInfo = pString;
	// var info
	*pVarInfo = 0x44;
	pVarInfo[4] = 1;
	wcscpy(reinterpret_cast<wchar_t*>(pVarInfo+6), L"VarFileInfo");
	pVarInfo[0x20] = 0x24;
	pVarInfo[0x22] = 4;
	wcscpy(reinterpret_cast<wchar_t*>(pVarInfo+0x26), L"Translation");
	*reinterpret_cast<DWORD*>(pVarInfo+0x40) = 0x04b00409;
	BYTE* pEnd = pVarInfo+0x44;
	*reinterpret_cast<WORD*>(szVersionInfo) = pEnd-szVersionInfo;
	UpdateResource(hRes, RT_VERSION, MAKEINTRESOURCE(1), 0, szVersionInfo, pEnd-szVersionInfo);
	if (a_hIcSmall && a_hIcMedium && a_hIcLarge)
	{
		BYTE aIconBits16[16*16*4+16*4+sizeof BITMAPINFOHEADER];
		GetIconBits(a_hIcSmall, aIconBits16, 16);
		UpdateResource(hRes, RT_ICON, MAKEINTRESOURCE(1), 0, aIconBits16, sizeof aIconBits16);
		BYTE aIconBits32[32*32*4+32*4+sizeof BITMAPINFOHEADER];
		GetIconBits(a_hIcMedium, aIconBits32, 32);
		UpdateResource(hRes, RT_ICON, MAKEINTRESOURCE(2), 0, aIconBits32, sizeof aIconBits32);
		BYTE aIconBits48[48*48*4+48*8+sizeof BITMAPINFOHEADER];
		GetIconBits(a_hIcLarge, aIconBits48, 48);
		UpdateResource(hRes, RT_ICON, MAKEINTRESOURCE(3), 0, aIconBits48, sizeof aIconBits48);

		GRPICONDIR tIcon;
		ZeroMemory(&tIcon, sizeof tIcon);
		tIcon.idType = 1;
		tIcon.idCount = 3;
		tIcon.idEntries[0].bWidth = 16;
		tIcon.idEntries[0].bHeight = 16;
		tIcon.idEntries[0].bColorCount = 0;
		//tIcon.idEntries[0].bReserved;
		tIcon.idEntries[0].wPlanes = 1;
		tIcon.idEntries[0].wBitCount = 32;
		tIcon.idEntries[0].dwBytesInRes = sizeof aIconBits16;
		tIcon.idEntries[0].nID = 1;
		tIcon.idEntries[1].bWidth = 32;
		tIcon.idEntries[1].bHeight = 32;
		tIcon.idEntries[1].bColorCount = 0;
		//tIcon.idEntries[1].bReserved;
		tIcon.idEntries[1].wPlanes = 1;
		tIcon.idEntries[1].wBitCount = 32;
		tIcon.idEntries[1].dwBytesInRes = sizeof aIconBits32;
		tIcon.idEntries[1].nID = 2;
		tIcon.idEntries[2].bWidth = 48;
		tIcon.idEntries[2].bHeight = 48;
		tIcon.idEntries[2].bColorCount = 0;
		//tIcon.idEntries[2].bReserved;
		tIcon.idEntries[2].wPlanes = 1;
		tIcon.idEntries[2].wBitCount = 32;
		tIcon.idEntries[2].dwBytesInRes = sizeof aIconBits48;
		tIcon.idEntries[2].nID = 3;
		UpdateResource(hRes, RT_GROUP_ICON, MAKEINTRESOURCE(1), 0, &tIcon, sizeof tIcon);
	}
	if (FALSE == EndUpdateResource(hRes, FALSE))
	{
		DeleteFile(szTemp);
		return false;
	}

	// move the file from %temp% to selected location
	HANDLE hDrp = CreateFile(szTemp, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL | FILE_FLAG_SEQUENTIAL_SCAN, NULL);
	if (hDrp == INVALID_HANDLE_VALUE)
	{
		DeleteFile(szTemp);
		return false;
	}
	ULONG nSize = GetFileSize(hDrp, NULL);
	CAutoVectorPtr<BYTE> aConfig(nSize ? new BYTE[nSize] : NULL);
	DWORD nRead;
	ReadFile(hDrp, aConfig, nSize, &nRead, NULL);
	CloseHandle(hDrp);
	DeleteFile(szTemp);
	CComPtr<IDataDstStream> pDst;
	a_pStorage->DstOpen(&pDst);
	if (pDst == NULL)
		return false;
	return SUCCEEDED(pDst->Write(nSize, aConfig));
}

#include <ContextHelpDlg.h>

class CDropletCreateDlg : 
	public Win32LangEx::CLangDialogImpl<CDropletCreateDlg>,
	public CDialogResize<CDropletCreateDlg>,
	public CContextHelpDlg<CDropletCreateDlg>,
	public IStorageFilterWindowCallback
{
public:
	CDropletCreateDlg(LCID a_tLocaleID, IStorageFilter** a_ppFilter, CComBSTR& a_bstrName, IEnumUnknowns* a_pFilters, IConfig* a_pOperationConfig, IConfig* a_pStorageConfig, bool* a_pOpenTarget, bool* a_pWaitForEnter, bool* a_pEnableGUI) :
		Win32LangEx::CLangDialogImpl<CDropletCreateDlg>(a_tLocaleID),
		CContextHelpDlg<CDropletCreateDlg>(_T("http://www.rw-designer.com/create-droplet"), NULL, EHLExeDir, NULL),
		m_ppFilter(a_ppFilter), m_bstrName(a_bstrName),
		m_pFilters(a_pFilters), m_pOperationConfig(a_pOperationConfig), m_pOpenTarget(a_pOpenTarget),
		m_pStorageConfig(a_pStorageConfig), m_pWaitForEnter(a_pWaitForEnter), m_pEnableGUI(a_pEnableGUI)
	{
	}
	~CDropletCreateDlg()
	{
	}

	enum { IDD = IDD_CREATEDROPLET };

	BEGIN_MSG_MAP(CDropletCreateDlg)
		CHAIN_MSG_MAP(CDialogResize<CDropletCreateDlg>)
		CHAIN_MSG_MAP(CContextHelpDlg<CDropletCreateDlg>)
		MESSAGE_HANDLER(WM_INITDIALOG, OnInitDialog)
		COMMAND_HANDLER(IDOK, BN_CLICKED, OnClickedOK)
		COMMAND_HANDLER(IDCANCEL, BN_CLICKED, OnClickedCancel)
	END_MSG_MAP()

	BEGIN_DLGRESIZE_MAP(CDropletCreateDlg)
		DLGRESIZE_CONTROL(IDC_CD_STORAGE, DLSZ_SIZE_Y|DLSZ_SIZE_X)
		DLGRESIZE_CONTROL(IDC_CD_TARGET_LABEL, DLSZ_MOVE_Y)
		DLGRESIZE_CONTROL(IDC_CD_TARGET, DLSZ_MOVE_Y|DLSZ_SIZE_X)
		DLGRESIZE_CONTROL(IDC_CD_OPENTARGET, DLSZ_MOVE_Y)
		DLGRESIZE_CONTROL(IDC_CD_WAITING, DLSZ_MOVE_Y)
		DLGRESIZE_CONTROL(IDC_CD_USERINTERFACE, DLSZ_MOVE_Y)
		DLGRESIZE_CONTROL(IDOK, DLSZ_MOVE_Y|DLSZ_MOVE_X)
		DLGRESIZE_CONTROL(IDCANCEL, DLSZ_MOVE_Y|DLSZ_MOVE_X)
		DLGRESIZE_CONTROL(IDHELP, DLSZ_MOVE_Y)
	END_DLGRESIZE_MAP()

	BEGIN_CTXHELP_MAP(CDropletCreateDlg)
		CTXHELP_CONTROL_RESOURCE(IDC_CD_TARGET, IDS_CD_TARGET_HELP)
		CTXHELP_CONTROL_RESOURCE(IDC_CD_OPENTARGET, IDS_CD_OPENTARGET_HELP)
		CTXHELP_CONTROL_RESOURCE(IDC_CD_WAITING, IDS_CD_WAITING_HELP)
		CTXHELP_CONTROL_RESOURCE(IDC_CD_USERINTERFACE, IDS_CD_USERINTERFACE_HELP)
	END_CTXHELP_MAP()

	bool DoModalPreTranslate(MSG const* a_pMsg)
	{
		if (!m_pWnd)
			return false;
		return m_pWnd->PreTranslateMessage(a_pMsg, TRUE) == S_OK || m_pWnd->PreTranslateMessage(a_pMsg, FALSE) == S_OK;
	}

	// IStorageFilterWindowCallback methods
public:
	STDMETHOD(QueryInterface)(const IID &,void **)
	{
		return E_NOTIMPL;
	}
	STDMETHOD_(ULONG, AddRef)()
	{
		return 1;
	}
	STDMETHOD_(ULONG, Release)()
	{
		return 0;
	}
	STDMETHOD(ForwardOK)()
	{
		if (S_OK == m_pWnd->FilterCreate(m_ppFilter))
		{
			if (m_pBatchSaveConfig)
			{
				CComBSTR bstrVal;
				GetDlgItemText(IDC_CD_TARGET, bstrVal.m_str);
				CComBSTR cCFGID_OUTPUTPATH(CFGID_OUTPUTPATH);
				m_pBatchSaveConfig->ItemValuesSet(1, &(cCFGID_OUTPUTPATH.m_str), CConfigValue(bstrVal));
			}
			*m_pWaitForEnter = IsDlgButtonChecked(IDC_CD_WAITING) == BST_CHECKED;
			*m_pEnableGUI = IsDlgButtonChecked(IDC_CD_USERINTERFACE) == BST_CHECKED;
			*m_pOpenTarget = IsDlgButtonChecked(IDC_CD_OPENTARGET) == BST_CHECKED;
			EndDialog(IDOK);
			return S_OK;
		}
		return S_FALSE;
	}
	STDMETHOD(ForwardCancel)()
	{
		EndDialog(IDCANCEL);
		return S_OK;
	}
	STDMETHOD(DefaultCommand)(ILocalizedString** UNREF(a_ppName), ILocalizedString** UNREF(a_ppDesc), GUID* UNREF(a_pIconID)) { return E_NOTIMPL; }
	STDMETHOD(DefaultCommandIcon)(ULONG UNREF(a_nSize), HICON* UNREF(a_phIcon)) { return E_NOTIMPL; }

	// handlers
public:
	LRESULT OnInitDialog(UINT a_uMsg, WPARAM a_wParam, LPARAM a_lParam, BOOL& a_bHandled)
	{
		if (m_pOperationConfig)
		{
			// hacking a bit -> if sequence is used and if last op is save to 
			CConfigValue cOpID;
			m_pOperationConfig->ItemValueGet(CComBSTR(CFGID_OPERATION), &cOpID);
			if (cOpID.TypeGet() == ECVTGUID && IsEqualGUID(cOpID, __uuidof(DocumentOperationSequence)))
			{
				OLECHAR sz[128];
				swprintf(sz, L"%s\\SeqSteps", CFGID_OPERATION);
				CConfigValue cSteps;
				m_pOperationConfig->ItemValueGet(CComBSTR(sz), &cSteps);
				if (cSteps.TypeGet() == ECVTInteger && cSteps.operator LONG() > 0)
				{
					swprintf(sz, L"%s\\SeqSteps\\%08x\\SeqOperation", CFGID_OPERATION, cSteps.operator LONG()-1);
					CConfigValue cLastOp;
					m_pOperationConfig->ItemValueGet(CComBSTR(sz), &cLastOp);
					if (cLastOp.TypeGet() == ECVTGUID && IsEqualGUID(cLastOp, BatchOperationManagerID))
					{
						m_pOperationConfig->SubConfigGet(CComBSTR(sz), &m_pBatchSaveConfig);
					}
				}
			}
		}
		if (m_pBatchSaveConfig)
		{
			CConfigValue cTemplate;
			m_pBatchSaveConfig->ItemValueGet(CComBSTR(CFGID_OUTPUTPATH), &cTemplate);
			SetDlgItemText(IDC_CD_TARGET, COLE2CT(cTemplate.operator BSTR()));
			if (wcsncmp(cTemplate.operator BSTR(), L"%BATCH%", 7) == 0)
				CheckDlgButton(IDC_CD_OPENTARGET, BST_CHECKED);
		}
		else
		{
			GetDlgItem(IDC_CD_TARGET).EnableWindow(FALSE);
			GetDlgItem(IDC_CD_TARGET_LABEL).EnableWindow(FALSE);
			GetDlgItem(IDC_CD_OPENTARGET).EnableWindow(FALSE);
		}
		if (*m_pWaitForEnter) CheckDlgButton(IDC_CD_WAITING, BST_CHECKED);
		if (*m_pEnableGUI) CheckDlgButton(IDC_CD_USERINTERFACE, BST_CHECKED);

		CWindow wnd = GetDlgItem(IDC_CD_STORAGE);
		RECT rc;
		wnd.GetWindowRect(&rc);
		ScreenToClient(&rc);
		CComPtr<IConfig> pCfg = m_pStorageConfig;
		CComPtr<IStorageFilterFactory> pFct;
		RWCoCreateInstance(pFct, __uuidof(StorageFilterFactoryFileSystem));
		if (pCfg == NULL)
			pFct->ContextConfigGetDefault(&pCfg);
		pFct->WindowCreate(m_bstrName, EFTCreateNew, m_hWnd, m_pFilters, pCfg, this, NULL, m_tLocaleID, &m_pWnd);
		RWHWND h = NULL;
		m_pWnd->Handle(&h);
		::MoveWindow(h, rc.left, rc.top, rc.right-rc.left, rc.bottom-rc.top, FALSE);
		::SetWindowLong(h, GWL_ID, IDC_CD_STORAGE);
		m_pWnd->Show(TRUE);
		::SetWindowPos(h, wnd, 0, 0, 0, 0, SWP_NOSIZE|SWP_NOMOVE|SWP_NOACTIVATE);
		wnd.DestroyWindow();
		DlgResize_Init();
		CenterWindow(GetParent());
		return 1;  // Let the system set the focus
	}
	LRESULT OnClickedOK(WORD a_wNotifyCode, WORD a_wID, HWND a_hWndCtl, BOOL& a_bHandled)
	{
		ForwardOK();
		return 0;
	}
	LRESULT OnClickedCancel(WORD a_wNotifyCode, WORD a_wID, HWND a_hWndCtl, BOOL& a_bHandled)
	{
		ForwardCancel();
		return 0;
	}

private:
	CComPtr<IStorageFilterWindow> m_pWnd;
	IStorageFilter** m_ppFilter;
	CComBSTR& m_bstrName;
	IEnumUnknowns* m_pFilters;
	CComPtr<IConfig> m_pOperationConfig;
	CComPtr<IConfig> m_pStorageConfig;
	CComPtr<IConfig> m_pBatchSaveConfig;
	bool* m_pWaitForEnter;
	bool* m_pEnableGUI;
	bool* m_pOpenTarget;
};


