
#pragma once

#include "EditTool.h"
#include <math.h>

//#include <boost/spirit.hpp>
//using namespace boost::spirit;


struct CEditToolDataShapeShift
{
	enum EShapeTool
	{
		ESTTangentPush = 0,
		ESTNormalPush,
		ESTRestore,
		ESTCollapse,
		ESTExpand,
		ESTTwirl,
		ESTTurbulence,
		ESTCount
	};

	CEditToolDataShapeShift() : eTool(ESTTangentPush), fSize(50.0f)
	{
	}

	HRESULT FromString(BSTR a_bstr)
	{
		int i = 0;
		float f = 50.0f;
		if (a_bstr && 2 == swscanf(a_bstr, L"%i,%f", &i, &f))
		{
			if (i >= 0 && i < ESTCount)
				eTool = static_cast<EShapeTool>(i);
			fSize = f;
		}
		return S_OK;
	}
	HRESULT ToString(BSTR* a_pbstr)
	{
		wchar_t szTmp[16];
		swprintf(szTmp, L"%i,%g", int(eTool), fSize);
		*a_pbstr = SysAllocString(szTmp);
		return S_OK;
	}

	EShapeTool eTool;
	float fSize;
};


#include "EditToolShapeShiftDlg.h"


class CEditToolShapeShift :
	public CComObjectRootEx<CComMultiThreadModel>,
	public CEditToolBase<>
{
public:
	CEditToolShapeShift() : m_fGamma(1.0f), m_eRasterMode(ERMSmooth), m_fTrackLength(0.0f), m_bMouseLeft(true), m_rcDirty(RECT_EMPTY), m_fLastPressure(0.0f), m_fDrawnPressure(0.0f)
	{
	}

	BEGIN_COM_MAP(CEditToolShapeShift)
		COM_INTERFACE_ENTRY(IRasterImageEditTool)
	END_COM_MAP()


	// IRasterImageEditTool methods
public:
	STDMETHOD(Reset)()
	{
		try
		{
			RECT rc = m_rcDirty;
			m_rcDirty = RECT_EMPTY;
			ULONG nX = 0;
			ULONG nY = 0;
			M_Window()->Size(&nX, &nY);
			m_fLastPressure = m_fTrackLength = m_fDrawnPressure = 0.0f;
			if (nX && nY)
			{
				m_aOffsets.Free();
				m_aSource.Free();
				//m_aCached.Free();
				m_aOffsets.Allocate(nX*nY);
				ZeroMemory(m_aOffsets.m_p, nX*nY*sizeof*m_aOffsets.m_p);
				m_aSource.Allocate(nX*nY);
				//m_aCached.Allocate(nX*nY);
				m_nSizeX = nX;
				m_nSizeY = nY;
				//m_rcCacheValid = RECT_EMPTY;
				M_Window()->GetImageTile(0, 0, nX, nY, nX, EITIContent, m_aSource);
			}
			if (rc.left < rc.right && rc.top < rc.bottom)
				M_Window()->RectangleChanged(&rc);
			return S_OK;
		}
		catch (...)
		{
			return E_UNEXPECTED;
		}
	}

	STDMETHOD(IsDirty)(RECT* a_pImageRect, BOOL* a_pOptimizeImageRect, RECT* a_pSelectionRect)
	{
		if (a_pImageRect)
			*a_pImageRect = m_rcDirty;
		if (a_pOptimizeImageRect)
			*a_pOptimizeImageRect = TRUE;
		if (a_pSelectionRect)
			*a_pSelectionRect = RECT_EMPTY;
		return m_rcDirty.left < m_rcDirty.right && m_rcDirty.top < m_rcDirty.bottom ? S_OK : S_FALSE;
	}
	STDMETHOD(GetImageTile)(ULONG a_nX, ULONG a_nY, ULONG a_nSizeX, ULONG a_nSizeY, ULONG a_nStride, TRasterImagePixel* a_pBuffer)
	{
		if (m_aOffsets.m_p == NULL || m_aSource.m_p == NULL || a_nX+a_nSizeX > m_nSizeX || a_nY+a_nSizeY > m_nSizeY)
			return M_Window()->GetImageTile(a_nX, a_nY, a_nSizeX, a_nSizeY, a_nStride, EITIBackground, a_pBuffer);

		RECT const rc = {a_nX, a_nY, a_nX+a_nSizeX, a_nY+a_nSizeY};
		SOffset const* pOff = m_aOffsets.m_p+m_nSizeX*a_nY+a_nX;
		static TRasterImagePixel const tEmpty = {0, 0, 0, 0};
		for (ULONG nY = rc.top; nY < rc.bottom; ++nY)
		{
			if (m_eRasterMode == ERMSmooth)
			{
				for (ULONG nX = rc.left; nX < rc.right; ++nX, ++a_pBuffer, ++pOff)
				{
					if (pOff->nDX&0xff)
					{
						int nDX = pOff->nDX>>8;
						int x1 = nDX+nX;
						if (x1 < -1 || x1 >= m_nSizeX)
						{
							*a_pBuffer = tEmpty;
						}
						else if (pOff->nDY&0xff)
						{
							int nDY = pOff->nDY>>8;
							int y1 = nDY+nY;
							if (y1 < -1 || y1 >= m_nSizeY)
							{
								*a_pBuffer = tEmpty;
							}
							else
							{
								int yo1 = y1*m_nSizeX;
								TRasterImagePixel const t1 = y1 == -1 || x1 == -1 ? tEmpty : m_aSource[yo1+x1];
								TRasterImagePixel const t2 = y1 == m_nSizeY-1 || x1 == -1 ? tEmpty : m_aSource[yo1+m_nSizeX+x1];
								TRasterImagePixel const t3 = y1 == -1 || x1 == m_nSizeX-1 ? tEmpty : m_aSource[yo1+x1+1];
								TRasterImagePixel const t4 = y1 == m_nSizeY-1 || x1 == m_nSizeX-1 ? tEmpty : m_aSource[yo1+m_nSizeX+x1+1];
								ULONG const nIY = pOff->nDY&0xff;
								ULONG const nFY = 0x100-nIY;
								ULONG const nIX = pOff->nDX&0xff;
								ULONG const nFX = 0x100-nIX;
								int nA = t1.bA*nFY*nFX + t2.bA*nIY*nFX + t3.bA*nFY*nIX + t4.bA*nIY*nIX;
								if (nA)
								{
									a_pBuffer->bR = (t1.bR*nFY*nFX*t1.bA + t2.bR*nIY*nFX*t2.bA + t3.bR*nFY*nIX*t3.bA + t4.bR*nIY*nIX*t4.bA)/nA;
									a_pBuffer->bG = (t1.bG*nFY*nFX*t1.bA + t2.bG*nIY*nFX*t2.bA + t3.bG*nFY*nIX*t3.bA + t4.bG*nIY*nIX*t4.bA)/nA;
									a_pBuffer->bB = (t1.bB*nFY*nFX*t1.bA + t2.bB*nIY*nFX*t2.bA + t3.bB*nFY*nIX*t3.bA + t4.bB*nIY*nIX*t4.bA)/nA;
									a_pBuffer->bA = nA>>16;
								}
								else
								{
									*a_pBuffer = tEmpty;
								}
							}
						}
						else
						{
							int y = nY+(pOff->nDY>>8);
							if (y < 0 || y >= m_nSizeY)
							{
								*a_pBuffer = tEmpty;
							}
							else
							{
								int yo = y*m_nSizeX;
								TRasterImagePixel const t1 = x1 == -1 ? tEmpty : m_aSource[yo+x1];
								TRasterImagePixel const t2 = x1 == m_nSizeX-1 ? tEmpty : m_aSource[yo+1+x1];
								ULONG const nIX = pOff->nDX&0xff;
								ULONG const nFX = 0x100-nIX;
								int nA = t1.bA*nFX + t2.bA*nIX;
								if (nA)
								{
									a_pBuffer->bR = (t1.bR*nFX*t1.bA + t2.bR*nIX*t2.bA)/nA;
									a_pBuffer->bG = (t1.bG*nFX*t1.bA + t2.bG*nIX*t2.bA)/nA;
									a_pBuffer->bB = (t1.bB*nFX*t1.bA + t2.bB*nIX*t2.bA)/nA;
									a_pBuffer->bA = nA>>8;
								}
								else
								{
									*a_pBuffer = tEmpty;
								}
							}
						}
					}
					else
					{
						int x = nX+(pOff->nDX>>8);
						if (x < 0 || x >= m_nSizeX)
						{
							*a_pBuffer = tEmpty;
						}
						else if (pOff->nDY&0xff)
						{
							int nDY = pOff->nDY>>8;
							int y1 = nDY+nY;
							if (y1 < -1 || y1 >= m_nSizeY)
							{
								*a_pBuffer = tEmpty;
							}
							else
							{
								int yo1 = y1*m_nSizeX;
								TRasterImagePixel const t1 = y1 == -1 ? tEmpty : m_aSource[yo1+x];
								TRasterImagePixel const t2 = y1 == m_nSizeY-1 ? tEmpty : m_aSource[yo1+m_nSizeX+x];
								ULONG const nIY = pOff->nDY&0xff;
								ULONG const nFY = 0x100-nIY;
								int nA = t1.bA*nFY + t2.bA*nIY;
								if (nA)
								{
									a_pBuffer->bR = (t1.bR*nFY*t1.bA + t2.bR*nIY*t2.bA)/nA;
									a_pBuffer->bG = (t1.bG*nFY*t1.bA + t2.bG*nIY*t2.bA)/nA;
									a_pBuffer->bB = (t1.bB*nFY*t1.bA + t2.bB*nIY*t2.bA)/nA;
									a_pBuffer->bA = nA>>8;
								}
								else
								{
									*a_pBuffer = tEmpty;
								}
							}
						}
						else
						{
							int y = nY+(pOff->nDY>>8);
							if (y < 0 || y >= m_nSizeY)
								*a_pBuffer = tEmpty;
							else
								*a_pBuffer = m_aSource[m_nSizeX*y+x];
						}
					}
				}
			}
			else
			{
				for (ULONG nX = rc.left; nX < rc.right; ++nX, ++a_pBuffer, ++pOff)
				{
					int const x = nX+((pOff->nDX+0x80)>>8);
					int const y = nY+((pOff->nDY+0x80)>>8);
					if (x < 0 || x >= m_nSizeX || y < 0 || y >= m_nSizeY)
						*a_pBuffer = tEmpty;
					else
						*a_pBuffer = m_aSource[m_nSizeX*y+x];
				}
			}
			a_pBuffer += a_nStride-a_nSizeX;
			pOff += m_nSizeX-a_nSizeX;
		}
		return S_OK;
	}

	STDMETHOD(PreTranslateMessage)(MSG const* a_pMsg)
	{
		return S_FALSE;
	}
	STDMETHOD(GetCursor)(EControlKeysState a_eKeysState, TPixelCoords const* a_tPos, HCURSOR* a_phCursor)
	{
		return E_NOTIMPL;
	}

	STDMETHOD(SetState)(ISharedState* a_pState)
	{
		CComBSTR bstrText;
		a_pState->ToText(&bstrText);
		m_cData.FromString(bstrText);
		M_Window()->ControlLinesChanged();
		return S_OK;
	}
	STDMETHOD(SetGlobals)(EBlendingMode a_eBlendingMode, ERasterizationMode a_eRasterizationMode, ECoordinatesMode a_eCoordinatesMode, EShapeFillMode a_eShapeFillMode, float a_fOutlineWidth, float a_fGammaOverride)
	{
		if (a_eRasterizationMode != m_eRasterMode || a_fGammaOverride != m_fGamma)
		{
			m_eRasterMode = a_eRasterizationMode;
			m_fGamma = a_fGammaOverride;
			if (m_rcDirty.left < m_rcDirty.right)
				M_Window()->RectangleChanged(&m_rcDirty);
		}
		return S_OK;
	}

	STDMETHOD(ProcessInputEvent)(EControlKeysState a_eKeysState, TPixelCoords const* a_pPos, TPixelCoords const* a_pPointerSize, float a_fNormalPressure, float a_fTangentPressure, float a_fOrientation, float a_fRotation, float a_fZ, DWORD* a_pMaxIdleTime)
	{
		if (a_pPos == NULL)
		{
			if (!m_bMouseLeft)
			{
				m_bMouseLeft = true;
				M_Window()->ControlLinesChanged();
			}
			return S_OK;
		}

		bool bPosChange = m_tLastPos.fX != a_pPos->fX || m_tLastPos.fY != a_pPos->fY || m_bMouseLeft;
		float const fSize = m_cData.fSize < 0.1f ? 0.1f : m_cData.fSize;
		m_bMouseLeft = false;
		if (m_fLastPressure == 0.0f)
		{
			if (a_fNormalPressure > 0.05f)
			{
				// start drawing
				m_fTrackLength = 0.0f;
				m_fLastPressure = a_fNormalPressure;
				m_tLastPos = *a_pPos;
				static TPixelCoords const tDir = {0.0f, 0.0f};
				PushPixels(*a_pPos, a_fNormalPressure, fSize, tDir);
			}
			else
			{
				m_tLastPos = *a_pPos;
			}
		}
		else
		{
			if (a_fNormalPressure < 0.05f)
			{
				// stop drawing
				//M_Window()->ApplyChanges();
				m_fLastPressure = 0.0f;
				m_tLastPos = *a_pPos;
			}
			else
			{
				// continue drawing
				while (true)
				{
					float fDX = a_pPos->fX-m_tLastPos.fX;
					float fDY = a_pPos->fY-m_tLastPos.fY;
					float fDP = a_fNormalPressure-m_fLastPressure;
					float fDist = sqrtf(fDX*fDX + fDY*fDY);
					if (fDist+m_fTrackLength < m_fDrawnPressure*fSize*0.25f)
					{
						m_tLastPos = *a_pPos;
						m_fTrackLength += fDist;
						break;
					}
					float fLen = m_fDrawnPressure*fSize*0.25f-m_fTrackLength;
					TPixelCoords tPoint =
					{
						m_tLastPos.fX + fDX*fLen/fDist,
						m_tLastPos.fY + fDY*fLen/fDist
					};
					float fPressure = m_fLastPressure + fDP*fLen/fDist;
					TPixelCoords tDir = {tPoint.fX-m_tLastPos.fX, tPoint.fY-m_tLastPos.fY};
					PushPixels(tPoint, fPressure, fSize, tDir);
					m_tLastPos = tPoint;
					m_fLastPressure = fPressure;
					m_fTrackLength = 0.0f;
				}
			}
		}
		if (bPosChange)
			M_Window()->ControlLinesChanged();
		return S_OK;
	}

	STDMETHOD(GetControlLines)(IEditToolControlLines* a_pLines)
	{
		if (!m_bMouseLeft)
		{
			int const nSteps = 12;
			float const fStep = 3.14159265f*2.0f/nSteps;
			TPixelCoords tCenter = m_tLastPos;
			float fRadius = (m_cData.fSize < 0.1f ? 0.1f : m_cData.fSize)*0.5f;
			a_pLines->MoveTo(tCenter.fX, tCenter.fY-fRadius);
			for (int i = 1; i < nSteps; ++i)
			{
				float fAngle = i*fStep;
				a_pLines->LineTo(tCenter.fX+fRadius*sinf(fAngle), tCenter.fY-fRadius*cosf(fAngle));
			}
			a_pLines->Close();
		}

		return S_OK;
	}

private:
	struct SOffset
	{
		int nDX; // fixed point 24:8 bits 
		int nDY;
	};

	struct SLinear
	{
		SLinear(TPixelCoords const& a_tDir) : m_tDir(a_tDir)
		{
		}

		void operator()(float fX, float fY, SOffset* pOff) const
		{
			float const fD = sqrtf(fX*fX + fY*fY);
			if (fD > 25.0f)
				return;
			pOff->nDX -= m_tDir.fX*10.0f*(25.0f-fD);
			pOff->nDY -= m_tDir.fY*10.0f*(25.0f-fD);
			//m_aOffsets[int(a_pPos->fX) + m_nSizeX*int(a_pPos->fY)].nDX += 0x100;
			//m_aOffsets[int(a_pPos->fX) + m_nSizeX*int(a_pPos->fY)].nDY += 0x80;
		}

	private:
		TPixelCoords const m_tDir;
	};
	struct SRadial
	{
		SRadial(float const a_fDir) : m_fDir(a_fDir)
		{
		}

		void operator()(float fX, float fY, SOffset* pOff) const
		{
			float const fD = sqrtf(fX*fX + fY*fY);
			if (fD > 25.0f)
				return;
			pOff->nDX -= fX*m_fDir*10.0f*(25.0f-fD);
			pOff->nDY -= fY*m_fDir*10.0f*(25.0f-fD);
			//m_aOffsets[int(a_pPos->fX) + m_nSizeX*int(a_pPos->fY)].nDX += 0x100;
			//m_aOffsets[int(a_pPos->fX) + m_nSizeX*int(a_pPos->fY)].nDY += 0x80;
		}

	private:
		float const m_fDir;
	};
	struct SRestore
	{
		void operator()(float fX, float fY, SOffset* pOff) const
		{
			float const fD = sqrtf(fX*fX + fY*fY);
			if (fD > 25.0f)
				return;
			pOff->nDX = pOff->nDX*fD/25.0f;
			pOff->nDY = pOff->nDY*fD/25.0f;
		}
	};

	template<class TShifter>
	void PushPixels(TShifter const& a_tShifter, TPixelCoords const& a_tPos, float a_fNormalPressure, float a_fSize)
	{
		int nX1 = floorf(a_tPos.fX-a_fSize);
		int nY1 = floorf(a_tPos.fY-a_fSize);
		int nX2 = ceilf(a_tPos.fX+a_fSize);
		int nY2 = ceilf(a_tPos.fY+a_fSize);
		if (nX1 < 0) nX1 = 0; else if (nX2 > m_nSizeX) nX2 = m_nSizeX;
		if (nY1 < 0) nY1 = 0; else if (nY2 > m_nSizeY) nY2 = m_nSizeY;
		AddDirtyRect(nX1,nY1, nX2, nY2);
		for (int y = nY1; y < nY2; ++y)
		{
			SOffset* pOff = m_aOffsets.m_p+m_nSizeX*y+nX1;
			for (int x = nX1; x < nX2; ++x, ++pOff)
			{
				a_tShifter(x-a_tPos.fX, y-a_tPos.fY, pOff);
			}
		}
	}

	void PushPixels(TPixelCoords const& a_tPos, float a_fNormalPressure, float a_fSize, TPixelCoords const& a_tDir)
	{
		m_fDrawnPressure = 1.0f;//m_cData.bSize ? a_fPressure : 1.0f;
		if (m_cData.eTool == CEditToolDataShapeShift::ESTTangentPush || m_cData.eTool == CEditToolDataShapeShift::ESTNormalPush)
		{
			if (fabsf(a_tDir.fX) <= 1e-3f && fabsf(a_tDir.fY) <= 1e-3f)
				return;
			TPixelCoords tDir = {m_cData.eTool == CEditToolDataShapeShift::ESTTangentPush ? a_tDir.fX : -a_tDir.fY, m_cData.eTool == CEditToolDataShapeShift::ESTTangentPush ? a_tDir.fY : a_tDir.fX};
			float f = sqrtf(tDir.fX*tDir.fX + tDir.fY*tDir.fY)*5.0f/m_cData.fSize;
			tDir.fX /= f;
			tDir.fY /= f;
			SLinear sLin(tDir);
			PushPixels(sLin, a_tPos, a_fNormalPressure, a_fSize);
		}
		else if (m_cData.eTool == CEditToolDataShapeShift::ESTCollapse || m_cData.eTool == CEditToolDataShapeShift::ESTExpand)
		{
			SRadial sRad(m_cData.eTool == CEditToolDataShapeShift::ESTCollapse ? -1.0f : 1.0f);
			PushPixels(sRad, a_tPos, a_fNormalPressure, a_fSize);
		}
		else if (m_cData.eTool == CEditToolDataShapeShift::ESTTwirl)
		{
		}
		else if (m_cData.eTool == CEditToolDataShapeShift::ESTTurbulence)
		{
		}
		else // ESTRestore
		{
			SRestore sRes;
			PushPixels(sRes, a_tPos, a_fNormalPressure, a_fSize);
		}
	}
	void AddDirtyRect(int a_nX1, int a_nY1, int a_nX2, int a_nY2)
	{
		if (m_rcDirty.left > a_nX1) m_rcDirty.left = a_nX1;
		if (m_rcDirty.top > a_nY1) m_rcDirty.top = a_nY1;
		if (m_rcDirty.right < a_nX2) m_rcDirty.right = a_nX2;
		if (m_rcDirty.bottom < a_nY2) m_rcDirty.bottom = a_nY2;
	}

private:
	ERasterizationMode m_eRasterMode;
	float m_fGamma;
	CEditToolDataShapeShift m_cData;

	RECT m_rcDirty;
	CAutoVectorPtr<SOffset> m_aOffsets;
	CAutoVectorPtr<TRasterImagePixel> m_aSource;
	//CAutoVectorPtr<TRasterImagePixel> m_aCached;
	//RECT m_rcCacheValid;
	ULONG m_nSizeX;
	ULONG m_nSizeY;

	TPixelCoords m_tLastPos;
	bool m_bMouseLeft;
	float m_fLastPressure;
	float m_fTrackLength;
	float m_fDrawnPressure;
};

