
#pragma once

bool MyCreateDirectory(TCHAR* a_pszPath)
{
	if (CreateDirectory(a_pszPath, NULL))
		return true;

	DWORD dwErr = GetLastError();
	if (dwErr == ERROR_ALREADY_EXISTS)
		return true;

	if (dwErr == ERROR_PATH_NOT_FOUND)
	{
		LPTSTR p = _tcsrchr(a_pszPath, _T('\\'));
		if (p == NULL || (p-a_pszPath) < 2)
			return false;
		*p = _T('\0');
		bool bRet = MyCreateDirectory(a_pszPath);
		*p = _T('\\');
		if (bRet && CreateDirectory(a_pszPath, NULL))
			return true;
	}
	return false;
}


class CApplicationInfo : public IClassFactory, public IApplicationInfo, public ILocalizedString
{
public:
	CApplicationInfo() : m_nPortable(S_FALSE)
	{
		m_szUserFolder[0] = m_szAppFolder[0] = m_szEmail[0] = m_szPassword[0] = m_szInstalled[0] = L'\0';
	}
	void Init()
	{
		GetModuleFileName(hInst, m_szAppFolder, itemsof(m_szAppFolder)-10);
		GetLongPathName(m_szAppFolder, m_szAppFolder, itemsof(m_szAppFolder)-10);
		m_szAppFolder[itemsof(m_szAppFolder)-12] = L'\\';
		LPOLESTR pEnd = wcsrchr(m_szAppFolder, L'\\');
		wcscpy(pEnd+1, L"AppData");
		ULONG nRes = GetFileAttributes(m_szAppFolder);
		if (nRes != 0xffffffff && (nRes&FILE_ATTRIBUTE_DIRECTORY))
		{
			m_nPortable = S_OK;
		}
		else
		{
			wcscpy(pEnd+1, L"portable.txt");
			if (GetFileAttributes(m_szAppFolder) != 0xffffffff)
				m_nPortable = S_OK;
		}
		*pEnd = L'\0';

		if (m_nPortable == S_OK)
		{
			wcscpy(m_szUserFolder, m_szAppFolder);
			wcscat(m_szUserFolder, L"\\AppData");
			if (!MyCreateDirectory(m_szUserFolder))
				m_nPortable = S_FALSE; // cannot create local AppData folder - cancel portable mode
		}

		if (m_nPortable == S_FALSE)
		{
			LPITEMIDLIST pIDList = NULL;
			if (S_OK == SHGetSpecialFolderLocation(NULL, CSIDL_APPDATA, &pIDList))
			{
				if (SHGetPathFromIDList(pIDList, m_szUserFolder))
				{
					size_t n = wcslen(m_szUserFolder);
					wcscpy(m_szUserFolder+(m_szUserFolder[n-1] == L'\\' ? n-1 : n), L"\\RealWorld\\" RWAPPID);
					MyCreateDirectory(m_szUserFolder);
				}
			}
			else if (S_OK == SHGetSpecialFolderLocation(NULL, CSIDL_MYDOCUMENTS, &pIDList))
			{
				
				if (SHGetPathFromIDList(pIDList, m_szUserFolder))
				{
					size_t n = wcslen(m_szUserFolder);
					wcscpy(m_szUserFolder+(m_szUserFolder[n-1] == L'\\' ? n-1 : n), L"\\RealWorld\\" RWAPPID);
					MyCreateDirectory(m_szUserFolder);
				}
			}
			else
			{
				wcscpy(m_szUserFolder, m_szAppFolder);
			}
			if (pIDList)
			{
				CComPtr<IMalloc> pMalloc;
				SHGetMalloc(&pMalloc);
				pMalloc->Free(pIDList);
			}
		}

		// load user / license settings
		OLECHAR szPersonal[MAX_PATH];
		swprintf(szPersonal, L"%s\\personal.ini", m_szUserFolder);
		HANDLE hPersonal = CreateFile(szPersonal, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL | FILE_FLAG_SEQUENTIAL_SCAN, NULL);
		if (hPersonal != INVALID_HANDLE_VALUE)
		{
			ULONG nSize = GetFileSize(hPersonal, NULL);
			CAutoVectorPtr<char> aSrc(new char[nSize+1]);
			ULONG nRead = 0;
			ReadFile(hPersonal, aSrc.m_p, nSize, &nRead, NULL);
			if (nRead == nSize)
			{
				aSrc[nSize] = '\0';
				CA2W str(aSrc[0] == char(0xef) && aSrc[1] == char(0xbb) && aSrc[2] == char(0xbf) ? aSrc.m_p+3 : aSrc.m_p, CP_UTF8);
				LPCWSTR psz = str;
				struct SIniEntry
				{
					LPCOLESTR pszKey;
					LPOLESTR pszDest;
					ULONG nMax;
				} s_aIniEntries[] =
				{
					{L"email", m_szEmail, sizeof(m_szEmail)/sizeof(*m_szEmail)},
					{L"password", m_szPassword, sizeof(m_szPassword)/sizeof(*m_szPassword)},
				};
				while (*psz)
				{
					LPCWSTR pLine = psz;
					while (*psz && *psz != L'\n' && *psz != L'\r') ++psz;
					if (pLine != psz)
					{
						for (SIniEntry* p = s_aIniEntries; p < s_aIniEntries+sizeof(s_aIniEntries)/sizeof(*s_aIniEntries); ++p)
						{
							int n = wcslen(p->pszKey);
							if (psz-pLine > n && _wcsnicmp(pLine, p->pszKey, n) == 0 && pLine[n] == L'=')
							{
								pLine += n+1;
								n = min(psz-pLine, int(p->nMax-1));
								wcsncpy(p->pszDest, pLine, n);
								p->pszDest[n] = L'\0';
								break;
							}
						}
					}
					while (*psz == L'\n' || *psz == L'\r') ++psz;
				}
			}

			CloseHandle(hPersonal);
		}
	}
	bool SavePersonalInfo()
	{
		OLECHAR szPersonal[MAX_PATH];
		swprintf(szPersonal, L"%s\\personal.ini", m_szUserFolder);
		HANDLE h = CreateFile(szPersonal, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL | FILE_FLAG_SEQUENTIAL_SCAN, NULL);
		if (h == INVALID_HANDLE_VALUE)
			return false;

		wchar_t sz[1024] = L"";
		swprintf(sz+wcslen(sz), L"email=%s\r\npassword=%s\r\n\r\n", m_szEmail, m_szPassword);
		char szUTF[2048] = {0xef, 0xbb, 0xbf, 0};
		int nLen = WideCharToMultiByte(CP_UTF8, 0, sz, -1, szUTF+3, itemsof(szUTF)-4, NULL, NULL);

		ULONG nWrt = 0;
		WriteFile(h, szUTF, nLen, &nWrt, NULL);
		CloseHandle(h);
		return nWrt == nLen;
	}
	LPCOLESTR M_AppFolder() const { return m_szAppFolder; }
	LPCOLESTR M_UserFolder() const { return m_szUserFolder; }

	bool CheckActivation()
	{
		return true;
	}

	// IUnknown methods
public:
	STDMETHOD_(ULONG, AddRef)()
	{
		return 2;
	}
	STDMETHOD_(ULONG, Release)()
	{
		return 1;
	}
	STDMETHOD(QueryInterface)(REFIID a_iid, void** a_pp)
	{
		try
		{
			if (IsEqualIID(a_iid, __uuidof(IApplicationInfo)))
			{
				*a_pp = reinterpret_cast<void*>(static_cast<IApplicationInfo*>(this));
				return S_OK;
			}
			else if (IsEqualIID(a_iid, __uuidof(ILocalizedString)))
			{
				*a_pp = reinterpret_cast<void*>(static_cast<ILocalizedString*>(this));
				return S_OK;
			}
			else if (IsEqualIID(a_iid, __uuidof(IClassFactory)))
			{
				*a_pp = reinterpret_cast<void*>(static_cast<IClassFactory*>(this));
				return S_OK;
			}
			else if (IsEqualIID(a_iid, __uuidof(IUnknown)))
			{
				*a_pp = reinterpret_cast<void*>(static_cast<IUnknown*>(static_cast<IApplicationInfo*>(this)));
				return S_OK;
			}
			else
			{
				*a_pp = NULL;
				return E_NOINTERFACE;
			}
		}
		catch (...)
		{
			return E_POINTER;
		}
	}

	// IClassFactory methods
public:
	STDMETHOD(CreateInstance)(IUnknown* pUnkOuter, REFIID riid, void** ppvObject)
	{
		if (pUnkOuter)
			return E_FAIL;
		return QueryInterface(riid, ppvObject);
	}
	STDMETHOD(LockServer)(BOOL fLock)
	{
		return S_OK;
	}

	// IDesignerAppInfo methods
public:
	STDMETHOD(Name)(ILocalizedString** a_ppName)
	{
		*a_ppName = this;
		AddRef();
		return S_OK;
	}
	STDMETHOD(Version)(ULONG* a_pVersion)
	{
		static int const iVer[4] = {RWVER0409_PRODUCTINFO};
		*a_pVersion = (iVer[0]<<16)|(iVer[1]<<8)|iVer[2];
		return S_OK;
	}
	STDMETHOD(Icon)(ULONG a_nSize, HICON* a_phIcon)
	{
		*a_phIcon = NULL;
		*a_phIcon = (HICON)LoadImage(hInst, MAKEINTRESOURCE(IDR_MAINFRAME), IMAGE_ICON, a_nSize, a_nSize, LR_DEFAULTCOLOR);
		return S_OK;
	}
	STDMETHOD(Identifier)(BSTR* a_pbstrAppID)
	{
		*a_pbstrAppID = NULL;
		*a_pbstrAppID = ::SysAllocString(RWAPPID);
		return S_OK;
	}

	STDMETHOD(Portable)()
	{
		return m_nPortable;
	}
	STDMETHOD(AppDataFolder)(BSTR* a_pbstrFolder)
	{
		*a_pbstrFolder = NULL;
		*a_pbstrFolder = ::SysAllocString(m_szUserFolder);
		return S_OK;
	}
	STDMETHOD(AppRootFolder)(BSTR* a_pbstrFolder)
	{
		*a_pbstrFolder = NULL;
		*a_pbstrFolder = ::SysAllocString(m_szAppFolder);
		return S_OK;
	}

	STDMETHOD(LicensingMode)(ELicensingMode* a_eLicensing)
	{
		if (a_eLicensing) *a_eLicensing = ELMDonate;
		return S_OK;
	}
	STDMETHOD(License)(BSTR* a_pbstrUserName, BSTR* a_pbstrOrganization, BSTR* a_pbstrSerial, ULONG* a_pAppLicCode, ULONG* a_pDaysInstalled)
	{
		try
		{
			if (a_pbstrUserName) *a_pbstrUserName = NULL;
			if (a_pbstrOrganization) *a_pbstrOrganization = NULL;
			if (a_pbstrSerial) *a_pbstrSerial = NULL;
			if (a_pAppLicCode) *a_pAppLicCode = 0;
			if (a_pDaysInstalled) *a_pDaysInstalled = 0;
			return S_FALSE;
		}
		catch (...)
		{
			return E_UNEXPECTED;
		}
	}
	STDMETHOD(Account)(BSTR* a_pbstrServer, BSTR* a_pbstrLogin, BSTR* a_pbstrPassHash)
	{
		try
		{
			if (a_pbstrServer) *a_pbstrServer = NULL;
			if (a_pbstrLogin) *a_pbstrLogin = NULL;
			if (a_pbstrPassHash) *a_pbstrPassHash = NULL;
			if (a_pbstrServer)
#ifdef _DEBUG
				*a_pbstrServer = SysAllocString(L"http://localhost/rwd/");
#else
				*a_pbstrServer = SysAllocString(L"http://www.rw-designer.com/");
#endif
			if (a_pbstrLogin && m_szEmail[0])
				*a_pbstrLogin = SysAllocString(m_szEmail);
			if (a_pbstrPassHash && m_szPassword[0])
			{
				CW2A str(m_szPassword, CP_UTF8);
				LPCSTR psz = str;
				MD5::SMD5 tPass;
				MD5::Compute(psz, strlen(psz), &tPass);
				wchar_t sz[33];
				swprintf(sz, L"%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x", tPass.val[0], tPass.val[1], tPass.val[2], tPass.val[3], tPass.val[4], tPass.val[5], tPass.val[6], tPass.val[7], tPass.val[8], tPass.val[9], tPass.val[10], tPass.val[11], tPass.val[12], tPass.val[13], tPass.val[14], tPass.val[15]);
				*a_pbstrPassHash = SysAllocString(sz);
			}
			return S_OK;
		}
		catch (...)
		{
			return E_UNEXPECTED;
		}
	}

	STDMETHOD(AboutBox)(RWHWND a_hWndParent, LCID a_tLocaleID)
	{
		try
		{
			return AppAboutBox(a_hWndParent, a_tLocaleID, this);
		}
		catch (...)
		{
			return E_UNEXPECTED;
		}
	}
	STDMETHOD(LicenseBox)(RWHWND a_hWndParent, LCID a_tLocaleID)
	{
		try
		{
			return AppLicenseBox(a_hWndParent, a_tLocaleID, this);
		}
		catch (...)
		{
			return E_UNEXPECTED;
		}
	}
	STDMETHOD(AccountBox)(RWHWND a_hWndParent, LCID a_tLocaleID)
	{
		return E_NOTIMPL;
	}
	STDMETHOD(AccountSet)(BSTR a_bstrLogin, BSTR a_bstrPassword)
	{
		static OLECHAR nothing[] = L"";
		if (a_bstrLogin == NULL)
			a_bstrLogin = nothing;
		if (a_bstrPassword == NULL)
			a_bstrPassword = nothing;
		wcsncpy(m_szEmail, a_bstrLogin, itemsof(m_szEmail));
		m_szEmail[itemsof(m_szEmail)-1] = L'\0';
		wcsncpy(m_szPassword, a_bstrPassword, itemsof(m_szPassword));
		m_szPassword[itemsof(m_szPassword)-1] = L'\0';
		return SavePersonalInfo() ? S_OK : S_FALSE;
	}
	STDMETHOD(AccountHash)(BSTR a_bstrPassword, BSTR* a_pbstrHash)
	{
		if (a_bstrPassword == NULL || a_pbstrHash == NULL)
			return E_FAIL;
		CW2A str(a_bstrPassword, CP_UTF8);
		LPCSTR psz = str;
		MD5::SMD5 tPass;
		MD5::Compute(psz, strlen(psz), &tPass);
		wchar_t sz[33];
		swprintf(sz, L"%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x", tPass.val[0], tPass.val[1], tPass.val[2], tPass.val[3], tPass.val[4], tPass.val[5], tPass.val[6], tPass.val[7], tPass.val[8], tPass.val[9], tPass.val[10], tPass.val[11], tPass.val[12], tPass.val[13], tPass.val[14], tPass.val[15]);
		*a_pbstrHash = SysAllocString(sz);
		return S_OK;
	}
	STDMETHOD(LicenseSet)(BSTR a_bstrName, BSTR a_bstrOrganization, BSTR a_bstrSerial)
	{
		return S_FALSE;
	}

	// ILocalizedString methods
public:
	STDMETHOD(Get)(BSTR* a_pbstrString)
	{
		return GetLocalized(GetThreadLocale(), a_pbstrString);
	}
	STDMETHOD(GetLocalized)(LCID a_tLCID, BSTR* a_pbstrString)
	{
		if (a_pbstrString == NULL)
			return E_POINTER;
		//*a_pbstrString = NULL;
		//wchar_t szBuffer[256] = L"";
		//Win32LangEx::LoadStringW(hInst, IDR_MAINFRAME, szBuffer, 256, LANGIDFROMLCID(a_tLCID));
		*a_pbstrString = CComBSTR(APP_MSGBOX_CAPTION/*szBuffer*/).Detach();
		return S_OK;
	}

private:
	OLECHAR m_szUserFolder[MAX_PATH];
	OLECHAR m_szAppFolder[MAX_PATH];
	HRESULT m_nPortable;
	OLECHAR m_szEmail[64];
	OLECHAR m_szPassword[64];
	OLECHAR m_szInstalled[64];
};


#ifndef CUSTOM_APPINFOOBJECT
CApplicationInfo g_sAppInfo;
HRESULT __stdcall GetClassObjectAppInfo(REFCLSID rclsid, REFIID riid, LPVOID* ppv)
{
	return g_sAppInfo.QueryInterface(riid, ppv);
}
#endif