
#pragma once

//take the liberty of using 1 for focused item - it used
// to be defined anyway, before MS removed it.
#ifndef TVIS_FOCUSED
#define TVIS_FOCUSED	1
#else
#if TVIS_FOCUSED != 1
#error TVIS_FOCUSED was assumed to be 1
#endif
#endif

#ifndef MST_TIMER_PERIOD
#define MST_TIMER_PERIOD	75		//ms
#endif


namespace WTL
{

class CMultiTreeCtrl : public CWindowImpl<CMultiTreeCtrl, CTreeViewCtrl>
{
public:
	typedef CSimpleArray<HTREEITEM> CTreeItemList;

	DECLARE_WND_SUPERCLASS(_T("MultiTreeCtrl"), WC_TREEVIEW);

	BEGIN_MSG_MAP(CMultiTreeCtrl)
		MESSAGE_HANDLER(WM_LBUTTONDOWN, OnButtonDown)
		//MESSAGE_HANDLER(WM_RBUTTONDOWN, OnButtonDown)
		MESSAGE_HANDLER(WM_SETFOCUS, OnSetFocus)
		MESSAGE_HANDLER(WM_KILLFOCUS, OnKillFocus)
		MESSAGE_HANDLER(WM_KEYDOWN, OnKeyDown)
		REFLECTED_NOTIFY_CODE_HANDLER(TVN_ITEMEXPANDING, OnItemExpanding)
	END_MSG_MAP()

	LRESULT OnButtonDown(UINT a_uMsg, WPARAM a_wParam, LPARAM a_lParam, BOOL& a_bHandled)
	{
		POINT tPt = {GET_X_LPARAM(a_lParam), GET_Y_LPARAM(a_lParam)};
		UINT nHF;
		HTREEITEM hItem = HitTest(tPt, &nHF);
		bool bBase = false;
		if (hItem)
		{
			//base always handles expanding items
			//(doesn't really mean much to right button, but pass anyway)
			bBase = (nHF & (TVHT_ONITEMBUTTON/*|TVHT_ONITEMINDENT*/)) != 0;
			if (!bBase && a_uMsg == WM_LBUTTONDOWN && (GetStyle() & TVS_CHECKBOXES))
			{
				//when the tree has check-boxes, the default handler makes
				// a quick selection of the clicked item, then re-selects
				// the previously selected item - to cause a sel-changed
				// notification.  Fortunately it doesn't affect the multi-
				// selection, so can pass on.
				bBase = (nHF & TVHT_ONITEMSTATEICON) != 0;
			}
		}
		if (bBase)
		{
			a_bHandled = FALSE;
			return 0;
		}

		if (!hItem || (nHF & (TVHT_ONITEMRIGHT|TVHT_NOWHERE|TVHT_ONITEMINDENT))) {
			//clicked in space, do rubber-banding
			DoBanding(a_uMsg == WM_LBUTTONDOWN, a_wParam, tPt);
			return 0;
		}

		ATLASSERT(nHF & (TVHT_ONITEM|TVHT_ONITEMSTATEICON));	//nothing else left

		DoPreSelection(hItem, a_uMsg == WM_LBUTTONDOWN, a_wParam);
		DoAction(hItem, a_uMsg == WM_LBUTTONDOWN, a_wParam, tPt);
		return 0;
	}

	LRESULT OnSetFocus(UINT a_uMsg, WPARAM a_wParam, LPARAM a_lParam, BOOL& a_bHandled)
	{
		CallWindowProc(m_pfnSuperWindowProc, m_hWnd, a_uMsg, a_wParam, a_lParam);
		//'emulated' selected items will remain greyed
		// if application gets covered
		HTREEITEM hItem = GetFirstSelectedItem();
		RECT rect;
		while (hItem)
		{
			GetItemRect(hItem, &rect, TRUE);
			InvalidateRect(&rect);
			hItem = GetNextSelectedItem(hItem);
		}
		return 0;
	}

	LRESULT OnKillFocus(UINT a_uMsg, WPARAM a_wParam, LPARAM a_lParam, BOOL& a_bHandled)
	{
		CallWindowProc(m_pfnSuperWindowProc, m_hWnd, a_uMsg, a_wParam, a_lParam);
		//'emulated' selected items may not get
		// refreshed to grey
		HTREEITEM hItem = GetFirstSelectedItem();
		RECT rect;
		while (hItem)
		{
			GetItemRect(hItem, &rect, TRUE);
			InvalidateRect(&rect);
			hItem = GetNextSelectedItem(hItem);
		}
		return 0;
	}

	LRESULT OnKeyDown(UINT a_uMsg, WPARAM a_wParam, LPARAM a_lParam, BOOL& a_bHandled)
	{
		bool bCtrl = (GetKeyState(VK_CONTROL) & 0x8000) != 0;
		bool bShift = (GetKeyState(VK_SHIFT) & 0x8000) != 0;

		bool bDir = false;
		HTREEITEM hSel = NULL;
		switch (a_wParam)
		{
		case VK_UP:
			bDir = true;
		case VK_DOWN:
			//common
			hSel = GetSelectedItem();
			if (!m_hSelect)
				m_hSelect = hSel;

			if (!bCtrl && !bShift)
			{
				m_hSelect = NULL;	//reset
				SelectAll(false);
			}
			break;
		case VK_SPACE:
			hSel = GetSelectedItem();
			if (hSel)
			{
				bool b = IsSelected(hSel);
				if (bCtrl)
					SetItemState(hSel, b ? 0 : TVIS_SELECTED, TVIS_SELECTED);
				else if (!b)
					SetItemState(hSel, TVIS_SELECTED, TVIS_SELECTED);
			}
			return 0;		//don't call base class (it would start a search on the char)
	//TODO: the text search isn't stopped by this ^.  It may be done in the TranslateMessage,
	// so would have to trap PreTranslateMessage to avoid it.  If 'space' selection is
	// required then need to implement - I'm not bothered.
		}

		CallWindowProc(m_pfnSuperWindowProc, m_hWnd, a_uMsg, a_wParam, a_lParam);
		if (!hSel || (!bCtrl && !bShift) )
			return 0;

		HTREEITEM hNext = bDir ? GetPrevVisibleItem(hSel) : GetNextVisibleItem(hSel);
		if (!hNext)
			hNext = hSel;
		if (bShift)
			SelectRange(m_hSelect, hNext, TRUE);
		else if (bCtrl)
			SetItemState(hNext, TVIS_FOCUSED, TVIS_FOCUSED);

		return 0;
	}

	LRESULT OnItemExpanding(int a_idCtrl, LPNMHDR a_pnmh, BOOL& a_bHandled)
	{
		a_bHandled = FALSE;

		LPNMTREEVIEW pNMTreeView = reinterpret_cast<LPNMTREEVIEW>(a_pnmh);
		if ((pNMTreeView->action == TVE_COLLAPSE) || (pNMTreeView->action == TVE_COLLAPSERESET))
		{
			//clear selection of children, it would be confusing otherwise
			// - the notifications can be over-ridden to stop the de-selection
			// if required
			//unfortunately, the parent window can't over-ride this functionality
			// because MFC gives this class first crack.  So if changes are required
			// a derived class will have to be used..
			ATLASSERT(pNMTreeView->itemNew.hItem);

			//if a descendent item has focus the parent will get selected as a
			// consequence of collapsing the tree, so preserve
			// (if the parent was already selected it will gain focus, but
			//  that's acceptable)
			bool bWasSel = IsSelected(pNMTreeView->itemNew.hItem);
			bool bWasFocus = SelectChildren(pNMTreeView->itemNew.hItem, false, true);
			if (bWasFocus && !bWasSel)
				CTreeViewCtrl::SelectItem(NULL);	//stop parent from gaining selection
		}

		return 0;
	}

	BOOL SetItemState(HTREEITEM hItem, UINT nState, UINT nStateMask)
	{
		HTREEITEM hFocus = GetSelectedItem();		//current focus
		bool bWasFocus = (hFocus == hItem);
		bool bFocusWasSel = hFocus && IsSelected(hFocus);	//selection state of current focus
		bool bWasSel = IsSelected(hItem);		//select state of acting item

		UINT nS = nState & ~TVIS_FOCUSED;
		UINT nSM = nStateMask & ~TVIS_FOCUSED;

		if (nStateMask & TVIS_FOCUSED)
		{
			//wanted to affect focus
			if (nState & TVIS_FOCUSED)
			{
				//wanted to set focus
				if (!bWasFocus && bFocusWasSel) {
					//because SelectItem would de-select the current 'real' selection
					// (the one with focus), need to make the tree ctrl think there is
					// no 'real' selection but still keep the the old item selected
					//it has to be done before the SelectItem call because
					// otherwise the TVN_SELCHANGING/ED notification handlers
					// wouldn't be able to get the proper list of selected items
					CTreeViewCtrl::SelectItem(NULL);	//will cause notify, but can be taken as focus change
					CTreeViewCtrl::SetItemState(hFocus, TVIS_SELECTED, TVIS_SELECTED);
					UpdateWindow();
				}
				if (!CTreeViewCtrl::SelectItem(hItem))	//set focus (will consequently select, if not already focused)
					return FALSE;
				if (nStateMask & TVIS_SELECTED)
				{
					//wanted to affect select state
					if (nState & TVIS_SELECTED)
					{
						//wanted to select, already done if wasn't focused
						if (!bWasFocus || bFocusWasSel)
						{
							nS &= ~TVIS_SELECTED;
							nSM &= ~TVIS_SELECTED;
						}
					}
					//else wanted to clear, base call will do that
				}
				else
				{
					//didn't want to affect select state
					if (!bWasSel)
					{
						//it wasn't previously selected, let base clear (correct)
						nS &= ~TVIS_SELECTED;
						nSM |= TVIS_SELECTED;
					}
					//else was already selected, no harm done
				}
			}
			else
			{
				//wanted to clear focus
				if (bWasFocus)
				{
					//it had the focus
					CTreeViewCtrl::SelectItem(NULL);	//clear focus
					if (!(nStateMask & TVIS_SELECTED))
					{
						//didn't want to affect select state
						if (bWasSel)
						{
							//it was selected, so restore
							ATLASSERT( !(nS & TVIS_SELECTED) );
							ATLASSERT( !(nSM & TVIS_SELECTED) );
							//set state here, to avoid double-notify
							CTreeViewCtrl::SetItemState(hItem, TVIS_SELECTED, TVIS_SELECTED);
							//let base do other states
						}
					}
					else if (nState & TVIS_SELECTED)
					{
						//wanted to select (but clear focus)
						if (bWasSel)
						{
							//if was selected, restore
							CTreeViewCtrl::SetItemState(hItem, TVIS_SELECTED, TVIS_SELECTED);
						}
						//don't want to notify, default did it
						nS &= ~TVIS_SELECTED;
						nSM &= ~TVIS_SELECTED;
					}
				}
			}
		}

		if (!nSM)
			return TRUE;	//no other states to alter

		if (nSM & TVIS_SELECTED)
		{
			//still need to alter selection state
			NMTREEVIEW nmtv;
			nmtv.hdr.hwndFrom = m_hWnd;
			nmtv.hdr.idFrom = ::GetDlgCtrlID(m_hWnd);
			nmtv.hdr.code = TVN_SELCHANGING;
			nmtv.itemOld.mask = nmtv.itemNew.mask = 0;
			nmtv.itemOld.hItem = nmtv.itemNew.hItem = NULL;
			TVITEM& item = (nS & TVIS_SELECTED) ? nmtv.itemNew : nmtv.itemOld;
			item.mask = TVIF_HANDLE|TVIF_PARAM;
			item.hItem = hItem;
			item.lParam = GetItemData(hItem);
			if (_SendNotify(&nmtv.hdr))
				return FALSE;	//sel-changing stopped
			ATLVERIFY( CTreeViewCtrl::SetItemState(hItem, nS, nSM) );
			nmtv.hdr.code = TVN_SELCHANGED;
			_SendNotify(&nmtv.hdr);
			nS &= ~TVIS_SELECTED;
			nSM &= ~TVIS_SELECTED;
		}
		if (!nSM)
			return TRUE;
		return CTreeViewCtrl::SetItemState(hItem, nS, nSM);
	}

	bool IsSelected(HTREEITEM a_hItem) const
	{
		return TVIS_SELECTED == (TVIS_SELECTED & CTreeViewCtrl::GetItemState(a_hItem, TVIS_SELECTED));
	}

	HTREEITEM GetFirstSelectedItem() const
	{
		HTREEITEM hItem = GetRootItem();
		while (hItem)
		{
			if (IsSelected(hItem))
				break;
			hItem = GetNextVisibleItem(hItem);
		}
		return hItem;
	}

	HTREEITEM GetNextSelectedItem(HTREEITEM a_hItem) const
	{
		a_hItem = GetNextVisibleItem(a_hItem);
		while (a_hItem)
		{
			if (IsSelected(a_hItem))
				break;
			a_hItem = GetNextVisibleItem(a_hItem);
		}
		return a_hItem;
	}

	void SelectAll(bool a_bSelect = true)
	{
		UINT nState = a_bSelect ? TVIS_SELECTED : 0;
		HTREEITEM hItem = GetRootItem();
		while (hItem)
		{
			if (IsSelected(hItem) != a_bSelect)
				SetItemState(hItem, nState, TVIS_SELECTED);
			hItem = GetNextVisibleItem(hItem);
		}
	}

	void SelectAllIgnore(bool a_bSelect, HTREEITEM a_hIgnore)
	{
		// special case to avoid multiple notifications for the same item
		UINT nState = a_bSelect ? TVIS_SELECTED : 0;
		HTREEITEM hItem = GetRootItem();
		while (hItem)
		{
			if (hItem != a_hIgnore && IsSelected(hItem) != a_bSelect)
				SetItemState(hItem, nState, TVIS_SELECTED);
			hItem = GetNextVisibleItem(hItem);
		}
	}

	bool SelectChildren(HTREEITEM a_hParent, bool a_bSelect = true, bool a_bRecurse = true)
	{
		UINT nS = a_bSelect ? TVIS_SELECTED : 0;

		bool bFocusWasInHere = false;

		HTREEITEM hItem = GetNextItem(a_hParent, TVGN_CHILD);
		while (hItem)
		{
			UINT nState = GetItemState(hItem, TVIS_SELECTED|TVIS_EXPANDED|TVIS_FOCUSED);
			if ((nState & TVIS_SELECTED) != nS)
				SetItemState(hItem, nS, TVIS_SELECTED);
			bFocusWasInHere |= (nState & TVIS_FOCUSED) != 0;
			if (a_bRecurse && (nState & TVIS_EXPANDED))
				bFocusWasInHere |= SelectChildren(hItem, a_bSelect, a_bRecurse);
			hItem = GetNextSiblingItem(hItem);
		}
		return bFocusWasInHere;
	}

	void GetSelectedList(CTreeItemList& a_cList) const
	{
		a_cList.RemoveAll();

		HTREEITEM hItem = GetFirstSelectedItem();
		while (hItem)
		{
			a_cList.Add(hItem);
			hItem = GetNextSelectedItem(hItem);
		}
	}

private:
	void DoPreSelection(HTREEITEM a_hItem, bool a_bLeft, UINT a_nFlags)
	{
		if (a_bLeft)
		{
			//if shift key down, select immediately
			if (a_nFlags & MK_SHIFT)
			{
				if (!m_hSelect)
					m_hSelect = GetSelectedItem();	//focus
				SelectRange(m_hSelect, a_hItem, !(a_nFlags & MK_CONTROL));
				SetItemState(a_hItem, TVIS_FOCUSED, TVIS_FOCUSED);	//focus changes to last clicked
			}
			else
			{
				if (!(a_nFlags & MK_CONTROL))
				{
					//if ctrl was down, then the selection is delayed until
					// mouse up, otherwise select the one item
					if (!IsSelected(a_hItem))
						SelectAllIgnore(false, a_hItem);
					SetItemState(a_hItem, TVIS_SELECTED|TVIS_FOCUSED, TVIS_SELECTED|TVIS_FOCUSED);
				}
				m_hSelect = NULL;	//reset when a non-shift operation occurs
			}
			return;
		}

		//right mouse
		if (a_nFlags & (MK_CONTROL|MK_SHIFT))
		{
			if (!(a_nFlags & MK_SHIFT))
				m_hSelect = a_hItem;
			return;		//do nothing if shift or ctrl
		}
		if (!IsSelected(a_hItem))
			SelectAllIgnore(FALSE, a_hItem);
		SetItemState(a_hItem, TVIS_SELECTED|TVIS_FOCUSED, TVIS_SELECTED|TVIS_FOCUSED);
	}

	void DoAction(HTREEITEM a_hItem, bool a_bLeft, UINT a_nFlags, POINT a_tPoint)
	{
		::SetCapture(m_hWnd);
		SIZE tSizeDrag = {::GetSystemMetrics(SM_CXDRAG), ::GetSystemMetrics(SM_CYDRAG)};

		MSG msg;
		UINT nDone = 0;

		while (!nDone && ::GetMessage(&msg, NULL, 0, 0))
		{

			if (::GetCapture() != m_hWnd)
				break;

			switch (msg.message)
			{
			case WM_MOUSEMOVE:
				if (abs(GET_X_LPARAM(msg.lParam) - a_tPoint.x) > tSizeDrag.cx ||
					abs(GET_Y_LPARAM(msg.lParam) - a_tPoint.y) > tSizeDrag.cy)
				{
					nDone = 2;
				}
				//because we exit loop, button up will still be dispatched
				// which means WM_CONTEXTMENU will be sent after TVN_BEGINRDRAG
				// - this is the same behaviour as original tree
				break;

			case WM_LBUTTONUP:
			case WM_RBUTTONUP:
				nDone = 1;
				break;

			default:
				::DispatchMessage(&msg);
				break;
			}
		}

		::ReleaseCapture();

		//construct tree notification info
		NMTREEVIEW nmtv;
		nmtv.hdr.hwndFrom = m_hWnd;
		nmtv.hdr.idFrom = ::GetDlgCtrlID(m_hWnd);
		nmtv.itemNew.mask = TVIF_HANDLE|TVIF_PARAM;
		nmtv.itemNew.hItem = a_hItem;
		nmtv.itemNew.lParam = GetItemData(a_hItem);
		DWORD dwStyle = GetStyle();

		if (nDone == 1)
		{
			//click
			if (!(a_nFlags & MK_SHIFT) && a_bLeft)
			{
				UINT nState = TVIS_SELECTED;
				if (a_nFlags & MK_CONTROL)
					nState ^= (GetItemState(a_hItem, TVIS_SELECTED) & TVIS_SELECTED);
				else
					SelectAllIgnore(FALSE, a_hItem);
				SetItemState(a_hItem, TVIS_FOCUSED|nState, TVIS_FOCUSED|TVIS_SELECTED);
			}
			if (::GetFocus() != m_hWnd)
				::SetFocus(m_hWnd);
			nmtv.hdr.code = a_bLeft ? NM_CLICK : NM_RCLICK;
			_SendNotify(&nmtv.hdr);
		}
		else if (nDone == 2)
		{
			//drag
			SetItemState(a_hItem, TVIS_FOCUSED|TVIS_SELECTED, TVIS_FOCUSED|TVIS_SELECTED);
			if (!(dwStyle & TVS_DISABLEDRAGDROP))
			{
				nmtv.hdr.code = a_bLeft ? TVN_BEGINDRAG : TVN_BEGINRDRAG;
				nmtv.ptDrag = a_tPoint;
				_SendNotify(&nmtv.hdr);
			}
		}
	}

	void DoBanding(bool a_bLeft, UINT a_nFlags, POINT a_tPoint)
	{
		if (::GetFocus() != m_hWnd)
			::SetFocus(m_hWnd);

		::SetCapture(m_hWnd);

		CTreeItemList list;
		if (a_nFlags & (MK_SHIFT|MK_CONTROL))
			GetSelectedList(list);

		CClientDC dc(m_hWnd);
		RECT rcCli;
		GetClientRect(&rcCli);

		MSG msg;
		bool bDone = false;
		POINT tPt;
		SIZE tSizeDrag = {::GetSystemMetrics(SM_CXDRAG), ::GetSystemMetrics(SM_CYDRAG)};
		bool bDrag = false;
		SIZE tSizeEdge = {1, 1};

		UINT nTimer = SetTimer(1, MST_TIMER_PERIOD, NULL);	//for scroll
		POINT tScrPt = {GetScrollPos(SB_HORZ), GetScrollPos(SB_VERT)};
		RECT rc = {0, 0, 0, 0};
		UINT h = 0;
		HTREEITEM hItem = GetRootItem();
		if (hItem)
		{
			GetItemRect(hItem, &rc, FALSE);
			tScrPt.y *= (h = rc.bottom-rc.top);		//this assumes equal height items
		}
		a_tPoint.x += tScrPt.x;
		a_tPoint.y += tScrPt.y;

		while (!bDone && ::GetMessage(&msg, NULL, 0, 0))
		{
			if (::GetCapture() != m_hWnd)
				break;

			switch (msg.message)
			{
				case WM_TIMER:
					tPt = msg.pt;
					ScreenToClient(&tPt);
					if (tPt.x >= rcCli.left && tPt.y >= rcCli.top && tPt.x < rcCli.right && tPt.y < rcCli.bottom)
					{
						::DispatchMessage(&msg);
						break;
					}
					msg.lParam = MAKELPARAM(tPt.x, tPt.y);
					//fall through to mousemove

				case WM_MOUSEMOVE:
					tPt.x = GET_X_LPARAM(msg.lParam);
					tPt.y = GET_Y_LPARAM(msg.lParam);
					if (!bDrag)
					{
						if (abs(tPt.x - a_tPoint.x) <= tSizeDrag.cx && abs(tPt.y - a_tPoint.y) <= tSizeDrag.cy)
							break;
						bDrag = true;
						if (!(a_nFlags & (MK_CONTROL|MK_SHIFT)))
							SelectAll(false);
						UpdateWindow();
						rc.top = rc.bottom = a_tPoint.x;
						rc.left = rc.right = a_tPoint.y;
						dc.DrawDragRect(&rc, tSizeEdge, NULL, tSizeEdge);
					}

					dc.DrawDragRect(&rc, tSizeEdge, NULL, tSizeEdge);	//delete

					if (tPt.y < rcCli.top)
						::SendMessage(m_hWnd, WM_VSCROLL, SB_LINEUP, 0);
					else if (tPt.y >= rcCli.bottom)
						::SendMessage(m_hWnd, WM_VSCROLL, SB_LINEDOWN, 0);
					if (tPt.x < rcCli.left)
						::SendMessage(m_hWnd, WM_HSCROLL, SB_LINELEFT, 0);
					else if (tPt.x >= rcCli.right)
						::SendMessage(m_hWnd, WM_HSCROLL, SB_LINERIGHT, 0);

					tScrPt.x = a_tPoint.x - GetScrollPos(SB_HORZ);
					tScrPt.y = a_tPoint.y - GetScrollPos(SB_VERT) * h;
					if (tPt.x > tScrPt.x)
					{
						rc.left = tScrPt.x;
						rc.right = tPt.x;
					}
					else
					{
						rc.left = tPt.x;
						rc.right = tScrPt.x;
					}
					if (tPt.y > tScrPt.y)
					{
						rc.top = tScrPt.y;
						rc.bottom = tPt.y;
					}
					else
					{
						rc.top = tPt.y;
						rc.bottom = tScrPt.y;
					}
					UpdateSelectionForRect(&rc, a_nFlags, list);
					dc.DrawDragRect(&rc, tSizeEdge, NULL, tSizeEdge);	//draw
					break;

				case WM_LBUTTONUP:
				case WM_RBUTTONUP:
					bDone = true;
					break;

				case WM_KEYDOWN:
					if (LOWORD(msg.wParam) == VK_ESCAPE)
					{
						SelectAll(false);
						bDone = true;
						break;
					}
					//dispatch

				default:
					::DispatchMessage(&msg);
					break;
			}
		}
		KillTimer(nTimer);
		::ReleaseCapture();

		if (bDrag)
			dc.DrawDragRect(&rc, tSizeEdge, NULL, tSizeEdge);
		else
			if (!(a_nFlags & (MK_CONTROL|MK_SHIFT)))
				SelectAll(false);

		//construct notification info
		NMHDR hdr;
		hdr.hwndFrom = m_hWnd;
		hdr.idFrom = ::GetDlgCtrlID(m_hWnd);
		hdr.code = a_bLeft ? NM_CLICK : NM_RCLICK;
		_SendNotify(&hdr);

		//when banding, make sure we receive WM_CONTEXTMENU
		// every time - which is what the listview ctrl does
		::DispatchMessage(&msg);
	}

	void SelectRange(HTREEITEM a_hFirst, HTREEITEM a_hLast, bool a_bOnly = true)
	{
		//locate (and select) either first or last
		// (so order is arbitrary)
		HTREEITEM hItem = GetRootItem();
		while (hItem)
		{
			if ((hItem == a_hFirst) || (hItem == a_hLast))
			{
				if (a_hFirst != a_hLast)
				{
					if (!IsSelected(hItem))
						SetItemState(hItem, TVIS_SELECTED, TVIS_SELECTED);
					hItem = GetNextVisibleItem(hItem);
				}
				break;
			}

			if (a_bOnly && IsSelected(hItem))
				SetItemState(hItem, 0, TVIS_SELECTED);
			hItem = GetNextVisibleItem(hItem);
		}
		//select rest of range
		while (hItem)
		{
			if (!IsSelected(hItem))
				SetItemState(hItem, TVIS_SELECTED, TVIS_SELECTED);
			if ((hItem == a_hFirst) || (hItem == a_hLast))
			{
				hItem = GetNextVisibleItem(hItem);
				break;
			}
			hItem = GetNextVisibleItem(hItem);
		}
		if (!a_bOnly)
			return;
		while (hItem)
		{
			if (IsSelected(hItem))
				SetItemState(hItem, 0, TVIS_SELECTED);
			hItem = GetNextVisibleItem(hItem);
		}
	}

	void UpdateSelectionForRect(LPCRECT a_pRect, UINT a_nFlags, CTreeItemList& a_list)
	{
		HTREEITEM hItem = GetRootItem();
		while (hItem)
		{
			bool bSel = IsSelected(hItem);
			RECT rc;
			GetItemRect(hItem, &rc, false);//m_bBandLabel);
			int iPos = a_list.Find(hItem);
			if (IntersectRect(&rc, &rc, a_pRect))
			{
				//item in rect
				if (!bSel && iPos == -1)
					SetItemState(hItem, TVIS_SELECTED, TVIS_SELECTED);
				else if ((a_nFlags & MK_CONTROL) && iPos != -1)
					SetItemState(hItem, 0, TVIS_SELECTED);
				else if ((a_nFlags & MK_SHIFT) && iPos != -1)
					a_list.RemoveAt(iPos);		//if shift and in rect, don't lock anymore
			}
			else
			{
				//item not in rect
				if (bSel && iPos == -1)
					SetItemState(hItem, 0, TVIS_SELECTED);
				else if (iPos != -1)
					SetItemState(hItem, TVIS_SELECTED, TVIS_SELECTED);
			}
			hItem = GetNextVisibleItem(hItem);
		}
		UpdateWindow();
	}

	LRESULT _SendNotify(LPNMHDR a_pNMHDR)
	{
		// - helper to distinguish between default control generated notifications
		//   and this classes emulated ones (so can tell if focus or select notify)
		ATLASSERT(::GetParent(m_hWnd));	//never expected this

		//BOOL b = m_bEmulated;
		//m_bEmulated = true;
		LRESULT lRes = ::SendMessage(::GetParent(m_hWnd), WM_NOTIFY, (WPARAM)a_pNMHDR->idFrom, (LPARAM)a_pNMHDR);
		//m_bEmulated = b;
		return lRes;
	}

private:
	HTREEITEM m_hSelect;	//for shift selection
};

}; // namespace