
#include <png.h>
#include <Win32LangEx.h>

typedef void (FNModifySplash)(BYTE* a_pRGBA, ULONG a_nX, ULONG a_nY);
FNModifySplash* g_pfnModifySplash = NULL;

typedef BOOL (WINAPI FNUpdateLayeredWindow)(HWND hWnd, HDC hdcDst, POINT *pptDst, SIZE *psize, HDC hdcSrc, POINT *pptSrc, COLORREF crKey, BLENDFUNCTION *pblend, DWORD dwFlags);
FNUpdateLayeredWindow* g_pfnUpdateLayeredWindow = NULL;

class CPNGDataSource
{
public:
	CPNGDataSource(void const* a_pData) : m_pData(reinterpret_cast<BYTE const*>(a_pData)), m_nActPos(0)
	{
	}

	operator png_voidp()
	{
		return reinterpret_cast<png_voidp>(this);
	}

	static void PNGAPI PngRWPtr(png_structp a_pPNGStruct, png_bytep a_pBuffer, png_size_t a_nBufferLen)
	{
		CPNGDataSource* pThis = reinterpret_cast<CPNGDataSource*>(png_get_io_ptr(a_pPNGStruct));
		CopyMemory(a_pBuffer, pThis->m_pData+pThis->m_nActPos, a_nBufferLen);
		pThis->m_nActPos += a_nBufferLen;
	}

private:
	BYTE const* m_pData;
	size_t m_nActPos;
};

HGDIOBJ g_hOldBitmap = NULL;
HDC g_hSplashDC = NULL;
int g_nSplashSizeX = 0;
int g_nSplashSizeY = 0;

static void LoadSplashImage(LPCTSTR a_pszResourceID)
{
	HRSRC hPNG = FindResource(hInst, a_pszResourceID, _T("RT_PNGIMAGE"));
	HGLOBAL hMem = LoadResource(hInst, hPNG);
	CPNGDataSource cSrc(LockResource(hMem));

	png_structp png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
	png_infop info_ptr = png_create_info_struct(png_ptr);
	png_set_read_fn(png_ptr, (void *)&cSrc, CPNGDataSource::PngRWPtr);

	png_read_info(png_ptr, info_ptr);

	png_uint_32 nWidth;
	png_uint_32 nHeight;
	int nBitDepth;
	int nColorType;
	int nInterlaceType;

	png_get_IHDR(png_ptr, info_ptr, &nWidth, &nHeight, &nBitDepth, &nColorType, &nInterlaceType, NULL, NULL);
	bool bAlphaChannel = (nColorType&PNG_COLOR_MASK_ALPHA) == PNG_COLOR_MASK_ALPHA;
	g_nSplashSizeX = nWidth;
	g_nSplashSizeY = nHeight;

	png_bytep* row_pointers = reinterpret_cast<png_bytep*>(_alloca(sizeof(png_bytep)*nHeight));

	int const nRowLen = bAlphaChannel ? (nWidth<<2) : ((nWidth*3+3)&0xfffffffc);
	BITMAPINFO tBI;
	ZeroMemory(&tBI, sizeof(tBI));
	tBI.bmiHeader.biSize = sizeof(tBI.bmiHeader);
	tBI.bmiHeader.biWidth = nWidth;
	tBI.bmiHeader.biHeight = -nHeight;
	tBI.bmiHeader.biPlanes = 1;
	tBI.bmiHeader.biBitCount = bAlphaChannel ? 32 : 24;
	tBI.bmiHeader.biCompression = BI_RGB;
	BYTE* pData = NULL;
	HDC hScreenDC = GetDC(NULL);
	g_hSplashDC = CreateCompatibleDC(hScreenDC);
	HBITMAP hBmpSrc = CreateDIBSection(g_hSplashDC, &tBI, DIB_RGB_COLORS, reinterpret_cast<void**>(&pData), NULL, 0);

	png_uint_32 i;
	for (i = 0; i < nHeight; i++)
	{
		row_pointers[i] = pData + i*nRowLen;
	}
	png_read_image(png_ptr, row_pointers);
	png_read_end(png_ptr, info_ptr);
	png_destroy_read_struct(&png_ptr, &info_ptr, NULL);
	if (g_pfnModifySplash)
		g_pfnModifySplash(pData, nWidth, nHeight);

	g_hOldBitmap = SelectObject(g_hSplashDC, hBmpSrc);
	ReleaseDC(NULL, hScreenDC);
	//GdiFlush();
}

static void DeleteSplashImage()
{
	if (g_hSplashDC)
	{
		HGDIOBJ hObj = SelectObject(g_hSplashDC, g_hOldBitmap);
		if (hObj) DeleteObject(hObj);
		DeleteDC(g_hSplashDC);
		g_hSplashDC = NULL;
	}
}

LRESULT CALLBACK WndProcOpaque(HWND a_hWnd, UINT a_uMsg, WPARAM a_wParam, LPARAM a_lParam)
{
	if (a_uMsg == WM_PAINT)
	{
		PAINTSTRUCT ps;
		HDC	hDC = BeginPaint(a_hWnd, &ps);
		BitBlt(hDC, 0, 0, g_nSplashSizeX, g_nSplashSizeY, g_hSplashDC, 0, 0, SRCCOPY);
		EndPaint(a_hWnd, &ps);
		return 0;
	}
	else
	{
		return DefWindowProc(a_hWnd, a_uMsg, a_wParam, a_lParam);
	}
}
//#ifdef _DEBUG
#define WS_EX_TOPMOST_RELEASE 0
//#else
//#define WS_EX_TOPMOST_RELEASE WS_EX_TOPMOST
//#endif

bool MessageLoopWithTimeout(DWORD a_dwTime)
{
	MSG msg;

	while (1)
	{
		DWORD dwRet = MsgWaitForMultipleObjects(0, NULL, FALSE, a_dwTime, QS_ALLINPUT);

		if (dwRet != WAIT_OBJECT_0)
			return true;

		// There is one or more window message available. Dispatch them
		while (PeekMessage(&msg,0,0,0,PM_NOREMOVE))
		{
			if (msg.message == WM_QUIT)
				return false;

			// check for unicode window so we call the appropriate functions
			BOOL bUnicode = ::IsWindowUnicode(msg.hwnd);
			BOOL bRet;

			if (bUnicode)
				bRet = ::GetMessageW(&msg, NULL, 0, 0);
			else
				bRet = ::GetMessageA(&msg, NULL, 0, 0);

			if (bRet > 0)
			{
				::TranslateMessage(&msg);

				if (bUnicode)
					::DispatchMessageW(&msg);
				else
					::DispatchMessageA(&msg);
			}
		}
	}
}

unsigned __stdcall Splash(void* pArguments)
{
	HMODULE hUser32dll = NULL;
	try
	{
		hUser32dll = LoadLibrary(_T("user32.dll"));
		g_pfnUpdateLayeredWindow = (FNUpdateLayeredWindow*)GetProcAddress(hUser32dll, "UpdateLayeredWindow");

		LoadSplashImage(MAKEINTRESOURCE(IDR_SPLASH_32BIT));

		WNDCLASSEX wcex;
		wcex.cbSize = sizeof(WNDCLASSEX); 
		wcex.style			= CS_OWNDC;
		wcex.lpfnWndProc	= (WNDPROC)(g_pfnUpdateLayeredWindow != NULL ? DefWindowProc : WndProcOpaque);
		wcex.cbClsExtra		= 0;
		wcex.cbWndExtra		= 0;
		wcex.hInstance		= hInst;
		wcex.hIcon			= 0;
		wcex.hCursor		= LoadCursor(NULL, IDC_WAIT);
		wcex.hbrBackground	= NULL;//(HBRUSH)(COLOR_WINDOW+1);
		wcex.lpszMenuName	= 0;
		wcex.lpszClassName	= _T("Splash");
		wcex.hIconSm		= 0;
		RegisterClassEx(&wcex);

		POINT tPos = {
			(GetSystemMetrics(SM_CXSCREEN)-g_nSplashSizeX)>>1,
			(GetSystemMetrics(SM_CYSCREEN)-g_nSplashSizeY)>>1
		};
		HWND hWnd = NULL;
		bool bMessageLoop = true;
		if (g_pfnUpdateLayeredWindow != NULL)
		{
			hWnd = CreateWindowEx(WS_EX_TOPMOST_RELEASE | WS_EX_TOOLWINDOW | WS_EX_LAYERED, _T("Splash"), _T("RWDesigner"), WS_POPUP,
				tPos.x, tPos.y, g_nSplashSizeX, g_nSplashSizeY, NULL, NULL, hInst, NULL);

			BLENDFUNCTION tBlend;
			tBlend.BlendOp = AC_SRC_OVER;
			tBlend.BlendFlags = 0;
			tBlend.SourceConstantAlpha = 0x1f;
			tBlend.AlphaFormat = AC_SRC_ALPHA;

			SIZE tSize = {g_nSplashSizeX, g_nSplashSizeY};
			POINT tSrcPt = {0, 0};
			//HDC hDCDst = GetDC(NULL);
			g_pfnUpdateLayeredWindow(hWnd, NULL, &tPos, &tSize, g_hSplashDC, &tSrcPt, 0, &tBlend, ULW_ALPHA);
			//ReleaseDC(hWnd, hDCDst);

			ShowWindow(hWnd, SW_SHOWNORMAL);

			DWORD dwStart = GetTickCount();
			int const nTickPerStep = 28;
			while (true)
			{
				DWORD dwTime = GetTickCount();
				int j = (dwTime-dwStart+(nTickPerStep>>1))/nTickPerStep;
				BYTE bAlpha = j < 7 ? 0x1f+j*0x20 : 0xff;
				if (tBlend.SourceConstantAlpha != bAlpha)
				{
					tBlend.SourceConstantAlpha = bAlpha;
					g_pfnUpdateLayeredWindow(hWnd, NULL, &tPos, &tSize, g_hSplashDC, &tSrcPt, 0, &tBlend, ULW_ALPHA);
					if (bAlpha == 0xff)
						break;
				}
				DWORD dwNext = dwStart+nTickPerStep*(j+1);
				if (!MessageLoopWithTimeout(dwNext-dwTime))
				{
					bMessageLoop = false;
					break;
				}
			}
		}
		else
		{
			hWnd = CreateWindowEx(WS_EX_TOPMOST_RELEASE | WS_EX_TOOLWINDOW, _T("Splash"), _T("RWDesigner"), WS_POPUP,
				tPos.x, tPos.y, g_nSplashSizeX, g_nSplashSizeY, NULL, NULL, hInst, NULL);
			ShowWindow(hWnd, SW_SHOWNORMAL);
		}

		if (bMessageLoop)
		{
			MSG msg;
			while (GetMessage(&msg, NULL, 0, 0)) 
			{
				TranslateMessage(&msg);
				DispatchMessage(&msg);
			}
		}
		DestroyWindow(hWnd);

		DeleteSplashImage();
		if (hUser32dll) FreeLibrary(hUser32dll);
		return 0;
	}
	catch (...)
	{
		DeleteSplashImage();
		if (hUser32dll) FreeLibrary(hUser32dll);
		return 1;
	}
}
